from nfstream import NFPlugin
import numpy as np


class SimpleTIG(NFPlugin):
    ''' SimpleTIG | A simple Traffic interaction Graph (TIG) feature extractor.

    From each packet the plugin extracts the packet size and the packet direction,
    the final feature is a signed packet size (neg size for src2dst, pos size for dst2src).
    
    Why simple: The plugin doesnt support any packet/node feature customization.
    
    Output Features: 
        1. `udps.simple_tig_adj`: Adjacency matrix of the graph. Shape of (`n_packets`, `n_packets`). 
        2. `udps.simple_tig_features`: Node features, a signed packet size. Shape of (`n_packets`, 1).
        
    Paper:
        "Accurate Decentralized Application Identification via Encrypted Traffic Analysis Using Graph Neural Networks."
        
        By:
            - Meng Shen, 
            - Jinpeng Zhang, 
            - Liehuang Zhu, 
            - Ke Xu, 
            - Xiaojiang Du.
    '''
    def __init__(self, n_packets=10, size_type='raw'):
        '''
        Args:
            `n_packets` (int): The number of packets to process per flow.
                flows with less than the desired `n_packets` will be padded with 
                unconnected zero-valued vertices to match `n_packets`.
            `size_type` (str): The layer of a packet to compute the size from. The options are 
                "raw", "ip", "transport" or "payload".
        '''
        self.n_packets = n_packets
        self.size_type = size_type
        
    def on_init(self, packet, flow):
        '''
        on_init
        '''
        flow.udps.simple_tig_values = []
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        '''on_update
        '''
        if flow.bidirectional_packets > self.n_packets:
            return
        
        size = self._packet_size_by_type(packet, self.size_type)
        dir = -1 if packet.direction == 0 else 1
        flow.udps.simple_tig_values.append(size * dir)
        
    def on_expire(self, flow):
        '''on_expire
        '''
        padding_amount_required = self.n_packets - len(flow.udps.simple_tig_values)
        graph = generate_simple_TIG(flow.udps.simple_tig_values, padding=padding_amount_required)
            
        flow.udps.simple_tig_adj = graph.adj.tolist()
        flow.udps.simple_tig_features = graph.features.tolist()
        
        del flow.udps.simple_tig_values
        
    def _packet_size_by_type(self, packet, size_type):
        if size_type == 'raw':
            return packet.raw_size
        elif size_type == 'ip':
            return packet.ip_size
        elif size_type == 'transport':
            return packet.transport_size
        elif size_type == 'payload':
            return packet.payload_size
        
        return None


def generate_simple_TIG(values, padding=0):
    nodes = len(values) + padding
    graph = Graph(
        nodes, 
        features=np.concatenate([
            np.array(values).reshape(-1, 1), 
            np.zeros(padding).reshape(-1, 1)
        ])
    )
    bursts = get_bursts(values)
    
    for b in bursts:
        graph.connect_inner_burst(b)
    
    if len(bursts) > 1:
        for b1, b2 in zip(bursts[:-1], bursts[1:]):
            graph.connect_bursts(b1, b2)
    
    return graph


def get_bursts(data, ids=None):
    if ids == None:
        ids = np.arange(len(data))
    bursts = []
    last_sign = np.sign(data[0])
    last_id_idx = 0
    burst = []
    for val in data:
        if np.sign(val) != last_sign:
            bursts.append(Burst(burst, ids[last_id_idx:last_id_idx+len(burst)]))
            last_id_idx += len(burst)
            burst = []
        burst.append(val)
        last_sign = np.sign(val)
    bursts.append(Burst(burst, ids[last_id_idx:last_id_idx+len(burst)]))
        
    return bursts


class Graph:
    def __init__(self, n, features=None, f=None):
        self.adj = np.zeros((n, n))
        if features is not None:
            self.features = features
        else:
            self.features = np.zeros((n, f))
    
    def connect(self, v1, v2):
        self.adj[v1, v2] = 1
        self.adj[v2, v1] = 1
        
    def connect_inner_burst(self, burst):
        if burst.length() == 1:
            return
        
        for node_id1, node_id2 in zip(burst.ids[:-1], burst.ids[1:]):
            self.connect(node_id1, node_id2)
            
    def connect_bursts(self, burst1, burst2):
        self.connect(burst1.first(), burst2.first())
        if burst1.length() == 1 and burst2.length() > 1:
            self.connect(burst1.first(), burst2.last())
        elif burst1.length() > 1 and burst2.length() == 1:
            self.connect(burst1.last(), burst2.first())
        else:
            self.connect(burst1.last(), burst2.last())


class Burst:
    def __init__(self, values, ids):
        self.values = values
        self.ids = ids
        
    def length(self):
        return len(self.values)
    
    def first(self):
        return self.ids[0]
    
    def last(self):
        return self.ids[-1]
    
    def __str__(self):
        return self.__repr__()
    
    def __repr__(self):
        return str(list(zip(self.values, self.ids)))


import subprocess
import csv
import glob
import os
from datetime import datetime



class spec:
    def __init__(self) -> None:
        pass
    
class spec_ip_address:
    def __init__(self, ip) -> None:
        self.ip = ip

    def __str__(self) -> str:
        return f'ip.addr=={self.ip}'
    
class spec_port:
    def __init__(self, port) -> None:
        self.port = port

    def __str__(self) -> str:
        return f'tcp.port=={self.port} or udp.port=={self.port}' 
    
class spec_protocol:
    def __init__(self, proto) -> None:
        self.proto = proto

    def __str__(self) -> str:
        return f'{self.proto}'
    
class spec_timestamp_larger_than:
    def __init__(self, epoch_time, units='nano') -> None:
        self.epoch_time = epoch_time
        if units == 'nano':
            self.epoch_time /= 1000

    def __str__(self) -> str:
        return f'frame.time_epoch>={self.epoch_time}'

class spec_timestamp_smaller_than:
    def __init__(self, epoch_time, units='nano') -> None:
        self.epoch_time = epoch_time +1
        if units == 'nano':
            self.epoch_time /= 1000

    def __str__(self) -> str:
        return f'frame.time_epoch<={self.epoch_time}'
    
class spec_timestamp_range:
    def __init__(self, start_epoch_time, end_epoch_time, units='nano') -> None:
        start = spec_timestamp_larger_than(start_epoch_time, units)
        end = spec_timestamp_smaller_than(end_epoch_time, units)
        self.spec = spec_and([start, end])

    def __str__(self) -> str:
        return str(self.spec)

class spec_and:
    def __init__(self, specs) -> None:
        self.specs = specs
        
    def __str__(self) -> str:
        return '('+' and '.join([str(spec) for spec in self.specs])+')'

class spec_or:
    def __init__(self, specs) -> None:
        self.specs = specs
        
    def __str__(self) -> str:
        return '('+' or '.join([str(spec) for spec in self.specs])+')'
    
class spec_5tuple:
    def __init__(self, ip1, port1, ip2, port2, protocol) -> None:
        self.specs = [
            spec_ip_address(ip1),
            spec_port(port1),
            spec_ip_address(ip2),
            spec_port(port2),
            spec_protocol(protocol)
        ]
    
    def __str__(self) -> str:
        return f'{self.protocol} and ip.addr=={self.ip1} and {self.protocol}.port=={self.port1} and ip.addr=={self.ip2} and {self.protocol}.port=={self.port2}'


def run_tshark_wspecs(input_filepath, output_filepath, tshark_location, filter, error_log_filepath='./tshark_error_log.txt', ):
    splitted_save = f'-w {output_filepath}'.split()
    command =  f'{tshark_location} -r {input_filepath} -2 -n -R'
    splitted_command = command.split() + [str(filter)] + splitted_save
    print(' '.join(splitted_command))
    with open(output_filepath, "w") as outfile:
        with open(error_log_filepath, "w") as error_log_file:
            proc = subprocess.run(
                splitted_command, 
                stdout = outfile, 
                stderr = error_log_file,
                check=True
            )
     


def current_time():
    now = datetime.now()
    current_time = now.strftime("%H:%M:%S")
    return current_time



if __name__ == '__main__':
    filepath = '../temp/DoH-Firefox84-NextDNS-1-pcap-format.pcap'
    output_path = '../temp/output_sess_ex.pcap'
    
    sess_1_spec = spec_and([
        spec_ip_address('192.168.1.105'),
        spec_port(61296),
        spec_ip_address('213.227.181.51'),
        spec_port(443),
        spec_protocol('tcp'),
        spec_timestamp_range(1610353294304, 1610353332681)
    ])
    sess_2_spec = spec_and([
        spec_ip_address('192.168.1.105'),
        spec_port(61393),
        spec_ip_address('213.227.181.51'),
        spec_port(443),
        spec_protocol('tcp'),
        spec_timestamp_larger_than(1610353294304),
        spec_timestamp_smaller_than(1610353758977),
    ])
    
    specs = spec_or([
       sess_1_spec, 
       sess_2_spec 
    ])
    
    # spec_5tuple('192.168.1.105', 61296, '213.227.181.51', 443, 'tcp')

    print(specs)
    
    run_tshark_wspecs(filepath, output_path, 'tshark', specs)


import unittest
import os
from nfstream import NFStreamer
import numpy as np
from EIMTC.plugins.n_pkts_byte_freq import NPacketsByteFrequency
from EIMTC.plugins.dns_counter import DNSCounter
from EIMTC.plugins.most_freq_payload_len_ratio import MostFreqPayloadLenRatio
from EIMTC.plugins.simple_tig import SimpleTIG
from EIMTC.plugins.small_pkt_payload_ratio import SmallPacketPayloadRatio
from EIMTC.plugins.res_req_diff_time import ResReqDiffTime
from EIMTC.plugins.pkt_rel_time import PacketRelativeTime
from EIMTC.plugins.clump_flows import Clump_Flow
from EIMTC.plugins.packets_size_interarrival_time import Packets_size_and_interarrival_time
from EIMTC.plugins.n_bytes import NBytes


dirname = os.path.dirname(__file__)
pcaps_dir = os.path.join(dirname, 'pcaps')

'''
    Testing NFStream Plugins
                            '''



class TestSimpleTIG(unittest.TestCase):

    def test_simpletig_2_pkts_with_size_raw(self):
        # Given
        plugins = [SimpleTIG(n_packets=2, size_type='raw')]
        pcap_filepath = os.path.join(pcaps_dir, 'tls_pkt_rel_time_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertSequenceEqual(flow_at_test.udps.simple_tig_adj, 
            [ # Adjacancy matrix
                [0,1],
                [1,0]
            ]
        )
        self.assertSequenceEqual(flow_at_test.udps.simple_tig_features, 
            [ # Features matrix
                [-98],
                [229],
            ]
        )
        
    def test_simpletig_10_pkts_with_size_raw(self):
        # Given
        plugins = [SimpleTIG(n_packets=10, size_type='raw')]
        pcap_filepath = os.path.join(pcaps_dir, 'tls_pkt_rel_time_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertSequenceEqual(flow_at_test.udps.simple_tig_adj, 
            [ # Adjacancy matrix
                [0,1,0,0,0,0,0,0,0,0],
                [1,0,1,0,0,0,0,0,0,0],
                [0,1,0,1,0,0,0,0,0,0],
                [0,0,1,0,1,0,0,0,0,0],
                [0,0,0,1,0,1,0,0,0,0],
                [0,0,0,0,1,0,1,0,0,0],
                [0,0,0,0,0,1,0,1,0,0],
                [0,0,0,0,0,0,1,0,0,0],
                [0,0,0,0,0,0,0,0,0,0], # Extra padding
                [0,0,0,0,0,0,0,0,0,0], # Extra padding
            ]
        )
        self.assertSequenceEqual(flow_at_test.udps.simple_tig_features, 
            [ # Features matrix
                [-98],
                [229],
                [-98],
                [229],
                [-98],
                [229],
                [-98],
                [229],
                [0], # Extra padding
                [0], # Extra padding
            ]
        )
        
    def test_simpletig_10_pkts_with_size_ip(self):
        # Given
        plugins = [SimpleTIG(n_packets=10, size_type='ip')]
        pcap_filepath = os.path.join(pcaps_dir, 'tls_req_res_diff_time_single_2.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertSequenceEqual(flow_at_test.udps.simple_tig_adj, 
            [ # Adjacancy matrix
                [0,1,1,0,0,0,0,0,0,0],
                [1,0,0,0,1,0,0,0,0,0],
                [1,0,0,1,0,1,0,0,0,0],
                [0,0,1,0,1,0,0,0,0,0],
                [0,1,0,1,0,0,1,0,0,0],
                [0,0,1,0,0,0,1,1,0,0],
                [0,0,0,0,1,1,0,0,0,1],
                [0,0,0,0,0,1,0,0,1,0],
                [0,0,0,0,0,0,0,1,0,1], 
                [0,0,0,0,0,0,1,0,1,0], 
            ]
        )
        self.assertSequenceEqual(flow_at_test.udps.simple_tig_features, 
            [ # Features matrix
                [-93],
                [-138],
                [75],
                [79],
                [146],
                [-89],
                [-138],
                [75],
                [76],
                [146],
            ]
        )



class TestNBytes(unittest.TestCase):

    def test_100_bytes_udp_dns_single_packet(self):
        # Given
        plugins = [NBytes(n=100)]
        pcap_filepath = os.path.join(pcaps_dir, 'dns_2_single_response_packet.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 100)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [
                183, 237, 129, 128, 0, 1, 0, 3, 0, 0, 0, 0, 3, 110, 97, 118,
                11, 115, 109, 97, 114, 116, 115, 99, 114, 101, 101, 110, 9, 109, 105, 99,
                114, 111, 115, 111, 102, 116, 3, 99, 111, 109, 0, 0, 1, 0, 1, 192,
                12, 0, 5, 0, 1, 0, 0, 12, 86, 0, 31, 10, 119, 100, 45, 112,
                114, 111, 100, 45, 115, 115, 14, 116, 114, 97, 102, 102, 105, 99, 109, 97,
                110, 97, 103, 101, 114, 3, 110, 101, 116, 0, 192, 59, 0, 5, 0, 1,
                0, 0, 0, 23
            ],
        )
        
    def test_200_bytes_udp_dns_single_packet(self):
        # Given
        plugins = [NBytes(n=200)]
        pcap_filepath = os.path.join(pcaps_dir, 'dns_2_single_response_packet.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 170)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 170 payload bytes
                183, 237, 129, 128, 0, 1, 0, 3, 0, 0, 0, 0, 3, 110, 97, 118,
                11, 115, 109, 97, 114, 116, 115, 99, 114, 101, 101, 110, 9, 109, 105, 99,
                114, 111, 115, 111, 102, 116, 3, 99, 111, 109, 0, 0, 1, 0, 1, 192,
                12, 0, 5, 0, 1, 0, 0, 12, 86, 0, 31, 10, 119, 100, 45, 112,
                114, 111, 100, 45, 115, 115, 14, 116, 114, 97, 102, 102, 105, 99, 109, 97,
                110, 97, 103, 101, 114, 3, 110, 101, 116, 0, 192, 59, 0, 5, 0, 1,
                0, 0, 0, 23, 0, 52, 23, 119, 100, 45, 112, 114, 111, 100, 45, 115,
                115, 45, 101, 117, 45, 119, 101, 115, 116, 45, 49, 45, 102, 101, 10, 119,
                101, 115, 116, 101, 117, 114, 111, 112, 101, 8, 99, 108, 111, 117, 100, 97,
                112, 112, 5, 97, 122, 117, 114, 101, 192, 38, 192, 102, 0, 1, 0, 1,
                0, 0, 0, 3, 0, 4, 51, 144, 113, 175, 
            ] + [ # padding 30 bytes
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, # 16 bytes
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,       # 14 bytes
            ], 
        )
        
    def test_184_bytes_udp_dhcpv6(self):
        # Given
        plugins = [NBytes(n=184)]
        pcap_filepath = os.path.join(pcaps_dir, 'dhcpv6_request_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 184)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 200 payload bytes
                4, 103, 247, 192, 0, 8, 0, 2, 0, 0, 0, 1, 0, 14, 0, 1,
                0, 1, 23, 81, 195, 34, 8, 0, 39, 60, 141, 201, 0, 3, 0, 40,
                14, 8, 0, 39, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 24,
                253, 45, 171, 140, 2, 37, 0, 0, 0, 0, 0, 0, 0, 0, 9, 148,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 0, 14, 0, 0, 1, 55,
                0, 8, 77, 83, 70, 84, 32, 53, 46, 48, 0, 6, 0, 6, 0, 24,
                0, 23, 0, 17, 1, 171, 159, 127, 0, 8, 0, 2, 0, 0, 0, 1,
                0, 14, 0, 1, 0, 1, 23, 81, 195, 34, 8, 0, 39, 60, 141, 201,
                0, 3, 0, 12, 14, 8, 0, 39, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 39, 0, 6, 0, 4, 87, 73, 78, 49, 0, 16, 0, 14, 0, 0,
                1, 55, 0, 8, 77, 83, 70, 84, 32, 53, 46, 48, 0, 6, 0, 8,
                0, 24, 0, 23, 0, 17, 0, 39,
            ]
        )
        
    def test_100_bytes_udp_skype_eth_padding(self):
        '''
        This is a test for a bug. Due to old limitations, the minimum len of
        a frame is 64 bytes, in case that the frame is shorter than 64 bytes
        the ethernet header is padding zeros at the END of the packet 
        (and not where the ethernet header is located in the packet structure).
        '''
        # Given
        plugins = [NBytes(n=100)]
        pcap_filepath = os.path.join(pcaps_dir, 'udp_skype_eth_padding_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 100)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 100 payload bytes
                172, 60, 93, 81, 25, 29, 184, 133, 29, 137, 173, 248, 90, 74, 67, 173,
                147, 12, 198, 146, 181, 13, 199, 111, 134, 3, 218, 74, 8, 168, 167, 32,
                201, 190, 29, 168, 194, 149, 52, 132, 237, 205, 88, 102, 166, 17, 14, 118,
                17, 22, 225, 242, 127, 17, 3, 19, 95, 154, 167, 208, 127, 43, 64, 92,
                9, 81, 71, 124, 194, 225, 77, 108, 9, 23, 168, 69, 69, 101, 136, 135,
                125, 88, 183, 120, 38, 183, 126, 186, 4, 227, 178, 151, 80, 18, 237, 70, 
                45, 252, 152, 182, 
            ]
        )
        
    def test_120_bytes_udp_skype_eth_padding(self):
        '''
        This is a test for a bug. Due to old limitations, the minimum len of
        a frame is 64 bytes, in case that the frame is shorter than 64 bytes
        the ethernet header is padding zeros at the END of the packet 
        (and not where the ethernet header is located in the packet structure).
        '''
        # Given
        plugins = [NBytes(n=120)]
        pcap_filepath = os.path.join(pcaps_dir, 'udp_skype_eth_padding_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 106)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 106 payload bytes
                172, 60, 93, 81, 25, 29, 184, 133, 29, 137, 173, 248, 90, 74, 67, 173,
                147, 12, 198, 146, 181, 13, 199, 111, 134, 3, 218, 74, 8, 168, 167, 32,
                201, 190, 29, 168, 194, 149, 52, 132, 237, 205, 88, 102, 166, 17, 14, 118,
                17, 22, 225, 242, 127, 17, 3, 19, 95, 154, 167, 208, 127, 43, 64, 92,
                9, 81, 71, 124, 194, 225, 77, 108, 9, 23, 168, 69, 69, 101, 136, 135,
                125, 88, 183, 120, 38, 183, 126, 186, 4, 227, 178, 151, 80, 18, 237, 70,
                45, 252, 152, 182, 61, 40, 105, 156, 184, 237,
            ] + [ # 14 padding bytes
                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, # 14 bytes
            ]
        )
        
    def test_200_bytes_udp_snmp_ipv6(self):
        # Given
        plugins = [NBytes(n=200)]
        pcap_filepath = os.path.join(pcaps_dir, 'snmp_ipv6_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 200)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 200 payload bytes
                48, 62, 2, 1, 3, 48, 17, 2, 4, 41, 205, 177, 122, 2, 3, 0,
                255, 207, 4, 1, 4, 2, 1, 3, 4, 16, 48, 14, 4, 0, 2, 1,
                0, 2, 1, 0, 4, 0, 4, 0, 4, 0, 48, 20, 4, 0, 4, 0,
                160, 14, 2, 4, 96, 186, 16, 246, 2, 1, 0, 2, 1, 0, 48, 0,
                48, 121, 2, 1, 3, 48, 17, 2, 4, 41, 205, 177, 122, 2, 3, 0,
                255, 207, 4, 1, 0, 2, 1, 3, 4, 38, 48, 36, 4, 20, 128, 0,
                79, 77, 177, 170, 220, 173, 188, 137, 175, 250, 17, 141, 189, 83, 130, 76,
                107, 5, 2, 1, 3, 2, 3, 1, 10, 29, 4, 0, 4, 0, 4, 0,
                48, 57, 4, 20, 128, 0, 79, 77, 177, 170, 220, 173, 188, 137, 175, 250,
                17, 141, 189, 83, 130, 76, 107, 5, 4, 0, 168, 31, 2, 4, 96, 186,
                16, 246, 2, 1, 0, 2, 1, 0, 48, 17, 48, 15, 6, 10, 43, 6,
                1, 6, 3, 15, 1, 1, 4, 0, 65, 1, 4, 48, 129, 144, 2, 1,
                3, 48, 17, 2, 4, 41, 205, 177,
            ],
        )

    def test_200_bytes_udp_snmp_ipv4(self):
        # Given
        plugins = [NBytes(n=200)]
        pcap_filepath = os.path.join(pcaps_dir, 'snmp_ipv4_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 200)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 200 payload bytes
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 64, 2,
                3, 2, 185, 46, 2, 1, 0, 2, 1, 0, 48, 51, 48, 15, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 2, 1, 5, 1, 5, 0, 48, 15, 6,
                11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 1, 1, 5, 0, 48, 15,
                6, 11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 5, 0,
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 64, 2,
                3, 2, 185, 47, 2, 1, 0, 2, 1, 0, 48, 51, 48, 15, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 2, 1, 5, 1, 5, 0, 48, 15, 6,
                11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 1, 1, 5, 0, 48, 15,
                6, 11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 5, 0,
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 64, 2,
                3, 2, 185, 48, 2, 1, 0, 2, 1, 0, 48, 51, 48, 15, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 2, 1, 5,
            ],
        )

    def test_784_bytes_udp_snmp_ipv4(self):
        '''
        This is a test for a bug. 
        Scapy dissector didnt dissect correctly SNMP 
        packet bytes.
        '''
        # Given
        plugins = [NBytes(n=784)]
        pcap_filepath = os.path.join(pcaps_dir, 'snmp_ipv4_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 784)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 784 payload bytes
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 64, 2,
                3, 2, 185, 46, 2, 1, 0, 2, 1, 0, 48, 51, 48, 15, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 2, 1, 5, 1, 5, 0, 48, 15, 6,
                11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 1, 1, 5, 0, 48, 15,
                6, 11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 5, 0,
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 64, 2,
                3, 2, 185, 47, 2, 1, 0, 2, 1, 0, 48, 51, 48, 15, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 2, 1, 5, 1, 5, 0, 48, 15, 6,
                11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 1, 1, 5, 0, 48, 15,
                6, 11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 5, 0,
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 64, 2,
                3, 2, 185, 48, 2, 1, 0, 2, 1, 0, 48, 51, 48, 15, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 2, 1, 5, 1, 5, 0, 48, 15, 6,
                11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 1, 1, 5, 0, 48, 15,
                6, 11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 5, 0,
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 64, 2,
                3, 2, 185, 49, 2, 1, 0, 2, 1, 0, 48, 51, 48, 15, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 2, 1, 5, 1, 5, 0, 48, 15, 6,
                11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 1, 1, 5, 0, 48, 15,
                6, 11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 5, 0,
                48, 130, 0, 90, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 162,
                130, 0, 75, 2, 3, 2, 185, 46, 2, 1, 0, 2, 1, 0, 48, 130,
                0, 60, 48, 130, 0, 16, 6, 11, 43, 6, 1, 2, 1, 25, 3, 2,
                1, 5, 1, 2, 1, 3, 48, 130, 0, 16, 6, 11, 43, 6, 1, 2,
                1, 25, 3, 5, 1, 1, 1, 2, 1, 3, 48, 130, 0, 16, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 4, 1, 161,
                48, 130, 0, 90, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 162,
                130, 0, 75, 2, 3, 2, 185, 47, 2, 1, 0, 2, 1, 0, 48, 130,
                0, 60, 48, 130, 0, 16, 6, 11, 43, 6, 1, 2, 1, 25, 3, 2,
                1, 5, 1, 2, 1, 3, 48, 130, 0, 16, 6, 11, 43, 6, 1, 2,
                1, 25, 3, 5, 1, 1, 1, 2, 1, 3, 48, 130, 0, 16, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 4, 1, 161,
                48, 130, 0, 90, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 162,
                130, 0, 75, 2, 3, 2, 185, 48, 2, 1, 0, 2, 1, 0, 48, 130,
                0, 60, 48, 130, 0, 16, 6, 11, 43, 6, 1, 2, 1, 25, 3, 2,
                1, 5, 1, 2, 1, 3, 48, 130, 0, 16, 6, 11, 43, 6, 1, 2,
                1, 25, 3, 5, 1, 1, 1, 2, 1, 3, 48, 130, 0, 16, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 4, 1, 161,
                48, 130, 0, 90, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 162,
                130, 0, 75, 2, 3, 2, 185, 49, 2, 1, 0, 2, 1, 0, 48, 130,
                0, 60, 48, 130, 0, 16, 6, 11, 43, 6, 1, 2, 1, 25, 3, 2,
                1, 5, 1, 2, 1, 3, 48, 130, 0, 16, 6, 11, 43, 6, 1, 2,
                1, 25, 3, 5, 1, 1, 1, 2, 1, 3, 48, 130, 0, 16, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 4, 1, 161,
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 64, 2,
                3, 2, 185, 50, 2, 1, 0, 2, 1, 0, 48, 51, 48, 15, 6, 11,
                43, 6, 1, 2, 1, 25, 3, 2, 1, 5, 1, 5, 0, 48, 15, 6,
                11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 1, 1, 5, 0, 48, 15,
                6, 11, 43, 6, 1, 2, 1, 25, 3, 5, 1, 2, 1, 5, 0,
                48, 77, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99,
            ],
        )
        
    def test_784_bytes_udp_snmp_ipv4_2(self):
        '''
        This is a test for a bug. 
        Scapy dissector didnt dissect correctly SNMP 
        packet bytes.
        '''
        # Given
        plugins = [NBytes(n=784)]
        pcap_filepath = os.path.join(pcaps_dir, 'snmp_ipv4_single_2.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 784)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 784 payload bytes
                48, 130, 0, 176, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 162,
                130, 0, 161, 2, 3, 2, 190, 139, 2, 1, 0, 2, 1, 0, 48, 130,
                0, 146, 48, 130, 0, 22, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2,
                1, 13, 1, 1, 4, 6, 84, 114, 97, 121, 32, 49, 48, 130, 0, 22,
                6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 2, 4, 6,
                84, 114, 97, 121, 32, 50, 48, 130, 0, 28, 6, 12, 43, 6, 1, 2,
                1, 43, 8, 2, 1, 13, 1, 3, 4, 12, 77, 97, 110, 117, 97, 108,
                32, 80, 97, 112, 101, 114, 48, 130, 0, 31, 6, 12, 43, 6, 1, 2,
                1, 43, 8, 2, 1, 13, 1, 4, 4, 15, 77, 97, 110, 117, 97, 108,
                32, 69, 110, 118, 101, 108, 111, 112, 101, 48, 130, 0, 23, 6, 12, 43,
                6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 5, 4, 7, 77, 80, 32,
                84, 114, 97, 121,
                48, 116, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 103, 2,
                3, 2, 190, 145, 2, 1, 0, 2, 1, 0, 48, 90, 48, 16, 6, 12,
                43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 6, 5, 0, 48, 16,
                6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 7, 5, 0,
                48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 8,
                5, 0, 48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13,
                1, 9, 5, 0, 48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2,
                1, 13, 1, 10, 5, 0,
                48, 130, 0, 176, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 162,
                130, 0, 161, 2, 3, 2, 190, 141, 2, 1, 0, 2, 1, 0, 48, 130,
                0, 146, 48, 130, 0, 22, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2,
                1, 13, 1, 1, 4, 6, 84, 114, 97, 121, 32, 49, 48, 130, 0, 22,
                6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 2, 4, 6,
                84, 114, 97, 121, 32, 50, 48, 130, 0, 28, 6, 12, 43, 6, 1, 2,
                1, 43, 8, 2, 1, 13, 1, 3, 4, 12, 77, 97, 110, 117, 97, 108,
                32, 80, 97, 112, 101, 114, 48, 130, 0, 31, 6, 12, 43, 6, 1, 2,
                1, 43, 8, 2, 1, 13, 1, 4, 4, 15, 77, 97, 110, 117, 97, 108,
                32, 69, 110, 118, 101, 108, 111, 112, 101, 48, 130, 0, 23, 6, 12, 43,
                6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 5, 4, 7, 77, 80, 32,
                84, 114, 97, 121,
                48, 116, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 103, 2,
                3, 2, 190, 147, 2, 1, 0, 2, 1, 0, 48, 90, 48, 16, 6, 12,
                43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 6, 5, 0, 48, 16,
                6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 7, 5, 0,
                48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 8,
                5, 0, 48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13,
                1, 9, 5, 0, 48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2,
                1, 13, 1, 10, 5, 0,
                48, 116, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 162, 103, 2,
                3, 2, 190, 143, 2, 1, 2, 2, 1, 1, 48, 90, 48, 16, 6, 12,
                43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 6, 5, 0, 48, 16,
                6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 7, 5, 0,
                48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13, 1, 8,
                5, 0, 48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 13,
                1, 9, 5, 0, 48, 16, 6, 12, 43, 6, 1, 2, 1, 43, 8, 2,
                1, 13, 1, 10, 5, 0,
                48, 80, 2, 1, 0, 4, 6, 112, 117, 98, 108, 105, 99, 160, 67, 2,
                3, 2, 190, 149, 2, 1, 0, 2, 1, 0, 48, 54, 48, 16, 6, 12,
                43, 6, 1, 2, 1, 43, 8, 2, 1, 3, 1, 0, 5, 0, 48, 16,
                6, 12, 43, 6, 1, 2, 1, 43, 8, 2, 1, 4, 1, 0, 5, 0,
                48, 16, 6, 12, 43, 6,
            ],
        )

    def test_100_bytes_tcp_tls_single_packet(self):
        # Given
        plugins = [NBytes(n=100)]
        pcap_filepath = os.path.join(pcaps_dir, 'tls_client_hello_single_packet.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 100)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 100 payload bytes
                22, 3, 1, 2, 0, 1, 0, 1, 252, 3, 3, 251, 155, 115, 218, 63,
                184, 40, 134, 145, 47, 235, 103, 119, 243, 70, 243, 216, 30, 130, 0, 27,
                249, 234, 186, 31, 2, 62, 63, 69, 107, 16, 11, 32, 234, 24, 167, 248,
                121, 83, 159, 124, 136, 56, 104, 182, 35, 253, 199, 120, 115, 130, 161, 42,
                62, 135, 150, 65, 113, 200, 150, 92, 180, 1, 218, 51, 0, 36, 19, 1,
                19, 3, 19, 2, 192, 43, 192, 47, 204, 169, 204, 168, 192, 44, 192, 48,
                192, 10, 192, 9
            ],
        )
        
    def test_16_bytes_tcp_ack_empty_single_packet(self):
        # Given
        plugins = [NBytes(n=16)]
        pcap_filepath = os.path.join(pcaps_dir, 'tcp_ack_single_packet.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 0)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # Empty payload of TCP ACK packet.
            ] + [ # 16 padding bytes.
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            ],
        )
        
    def test_16_bytes_tcp_syn_empty_single_packet(self):
        # Given
        plugins = [NBytes(n=16)]
        pcap_filepath = os.path.join(pcaps_dir, 'tcp_syn_single_packet.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 0)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # Empty payload of TCP SYN packet.
            ] + [ # 16 padding bytes.
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            ],
        )
        
    def test_300_bytes_tcp_http(self):
        # Given
        plugins = [NBytes(n=300)]
        pcap_filepath = os.path.join(pcaps_dir, 'http_most_freq_payload_ratio_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamers
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.n_bytes_counted, 300)
        self.assertSequenceEqual(list(flow_at_test.udps.n_bytes), 
            [ # 300 payload bytes
                71, 69, 84, 32, 47, 115, 117, 99, 99, 101, 115, 115, 46, 116, 120, 116,
                32, 72, 84, 84, 80, 47, 49, 46, 49, 13, 10, 72, 111, 115, 116, 58,
                32, 100, 101, 116, 101, 99, 116, 112, 111, 114, 116, 97, 108, 46, 102, 105,
                114, 101, 102, 111, 120, 46, 99, 111, 109, 13, 10, 85, 115, 101, 114, 45,
                65, 103, 101, 110, 116, 58, 32, 77, 111, 122, 105, 108, 108, 97, 47, 53,
                46, 48, 32, 40, 87, 105, 110, 100, 111, 119, 115, 32, 78, 84, 32, 49,
                48, 46, 48, 59, 32, 87, 105, 110, 54, 52, 59, 32, 120, 54, 52, 59,
                32, 114, 118, 58, 56, 52, 46, 48, 41, 32, 71, 101, 99, 107, 111, 47,
                50, 48, 49, 48, 48, 49, 48, 49, 32, 70, 105, 114, 101, 102, 111, 120,
                47, 56, 52, 46, 48, 13, 10, 65, 99, 99, 101, 112, 116, 58, 32, 42,
                47, 42, 13, 10, 65, 99, 99, 101, 112, 116, 45, 76, 97, 110, 103, 117,
                97, 103, 101, 58, 32, 101, 110, 45, 85, 83, 44, 101, 110, 59, 113, 61,
                48, 46, 53, 13, 10, 65, 99, 99, 101, 112, 116, 45, 69, 110, 99, 111,
                100, 105, 110, 103, 58, 32, 103, 122, 105, 112, 44, 32, 100, 101, 102, 108,
                97, 116, 101, 13, 10, 67, 97, 99, 104, 101, 45, 67, 111, 110, 116, 114,
                111, 108, 58, 32, 110, 111, 45, 99, 97, 99, 104, 101, 13, 10, 80, 114,
                97, 103, 109, 97, 58, 32, 110, 111, 45, 99, 97, 99, 104, 101, 13, 10,
                67, 111, 110, 110, 101, 99, 116, 105, 111, 110, 58, 32, 107, 101, 101, 112,
                45, 97, 108, 105, 118, 101, 13, 10, 13, 10, 72, 84
            ],
        )



class TestClumpsPlugin(unittest.TestCase):
    def setUp(self):
        self.plugins = [Clump_Flow()]

    def test_clump_flow_1(self):
            # Given
        pcap_filepath = os.path.join(pcaps_dir, 'tls_pkt_rel_time_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=self.plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        
        #src2dst
        self.assertEqual(flow_at_test.udps.src2dst_max_clumps_len, 98)
        self.assertEqual(flow_at_test.udps.src2dst_min_clumps_len, 98)
        self.assertEqual(flow_at_test.udps.src2dst_mean_clumps_len, 98)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clumps_len, 0)
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clumps_len, 0)
        self.assertEqual(flow_at_test.udps.src2dst_variance_clumps_len, 0)
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clumps_len, 0)
        
        self.assertEqual(flow_at_test.udps.src2dst_max_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.src2dst_min_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.src2dst_mean_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clumps_size, 0)
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clumps_size, 0)
        self.assertEqual(flow_at_test.udps.src2dst_variance_clumps_size, 0)
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clumps_size, 0)

        self.assertEqual(flow_at_test.udps.src2dst_max_clumps_bytes_per_packet, 98)
        self.assertEqual(flow_at_test.udps.src2dst_min_clumps_bytes_per_packet, 98)
        self.assertEqual(flow_at_test.udps.src2dst_mean_clumps_bytes_per_packet, 98)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clumps_bytes_per_packet, 0)
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clumps_bytes_per_packet, 0)
        self.assertEqual(flow_at_test.udps.src2dst_variance_clumps_bytes_per_packet, 0)
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clumps_bytes_per_packet, 0)

        self.assertEqual(flow_at_test.udps.src2dst_max_clumps_interarrival_time, 44952.0)
        self.assertEqual(flow_at_test.udps.src2dst_min_clumps_interarrival_time, 0.0)
        self.assertEqual(flow_at_test.udps.src2dst_mean_clumps_interarrival_time, 33690.0)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clumps_interarrival_time, 22460.031715620233)
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clumps_interarrival_time, -1.1546907587693247)
        self.assertEqual(flow_at_test.udps.src2dst_variance_clumps_interarrival_time, 504453024.6666667)
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clumps_interarrival_time, -0.666674206180156)

        self.assertEqual(flow_at_test.udps.src2dst_max_clump_len, [98.0,98.0,98.0,98.0])
        self.assertEqual(flow_at_test.udps.src2dst_min_clump_len, [98.0,98.0,98.0,98.0])
        self.assertEqual(flow_at_test.udps.src2dst_mean_clump_len, [98.0,98.0,98.0,98.0])
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clump_len, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clump_len, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_variance_clump_len, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clump_len, [0.0,0.0,0.0,0.0])

        self.assertEqual(flow_at_test.udps.src2dst_max_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_min_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_mean_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_variance_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clump_interarrival_time, [0.0,0.0,0.0,0.0])

        #dst2src
        self.assertEqual(flow_at_test.udps.dst2src_max_clumps_len, 229)
        self.assertEqual(flow_at_test.udps.dst2src_min_clumps_len, 229)
        self.assertEqual(flow_at_test.udps.dst2src_mean_clumps_len, 229)
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clumps_len, 0)
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clumps_len, 0)
        self.assertEqual(flow_at_test.udps.dst2src_variance_clumps_len, 0)
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clumps_len, 0)

        self.assertEqual(flow_at_test.udps.dst2src_max_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.dst2src_min_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.dst2src_mean_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clumps_size, 0)
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clumps_size, 0)
        self.assertEqual(flow_at_test.udps.dst2src_variance_clumps_size, 0)
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clumps_size, 0)

        self.assertEqual(flow_at_test.udps.dst2src_max_clumps_bytes_per_packet, 229)
        self.assertEqual(flow_at_test.udps.dst2src_min_clumps_bytes_per_packet, 229)
        self.assertEqual(flow_at_test.udps.dst2src_mean_clumps_bytes_per_packet, 229)
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clumps_bytes_per_packet, 0)
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clumps_bytes_per_packet, 0)
        self.assertEqual(flow_at_test.udps.dst2src_variance_clumps_bytes_per_packet, 0)
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clumps_bytes_per_packet, 0)

        self.assertEqual(flow_at_test.udps.dst2src_max_clumps_interarrival_time, 132.0)
        self.assertEqual(flow_at_test.udps.dst2src_min_clumps_interarrival_time, 63.0)
        self.assertEqual(flow_at_test.udps.dst2src_mean_clumps_interarrival_time, 82.5)
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clumps_interarrival_time, 33.090784215548595)
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clumps_interarrival_time, 1.1357582101135595)
        self.assertEqual(flow_at_test.udps.dst2src_variance_clumps_interarrival_time, 1095.0)
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clumps_interarrival_time, -0.6812084818915367)

        self.assertEqual(flow_at_test.udps.dst2src_max_clump_len, [229.0,229.0,229.0,229.0])
        self.assertEqual(flow_at_test.udps.dst2src_min_clump_len, [229.0,229.0,229.0,229.0])
        self.assertEqual(flow_at_test.udps.dst2src_mean_clump_len, [229.0,229.0,229.0,229.0])
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clump_len, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clump_len, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_variance_clump_len, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clump_len, [0.0,0.0,0.0,0.0])

        self.assertEqual(flow_at_test.udps.dst2src_max_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_min_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_mean_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_variance_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clump_interarrival_time, [0.0,0.0,0.0,0.0])

        #bidirectional
        self.assertEqual(flow_at_test.udps.max_clumps_len, 229)
        self.assertEqual(flow_at_test.udps.min_clumps_len,98)
        self.assertEqual(flow_at_test.udps.mean_clumps_len, 163.5)
        self.assertEqual(flow_at_test.udps.stddev_clumps_len,70.02244538105519)
        self.assertEqual(flow_at_test.udps.skewness_clumps_len,0)
        self.assertEqual(flow_at_test.udps.variance_clumps_len,4903.142857142857)
        self.assertEqual(flow_at_test.udps.kurtosis_clumps_len,-2.0)

        self.assertEqual(flow_at_test.udps.max_clumps_size,1)
        self.assertEqual(flow_at_test.udps.min_clumps_size,1)
        self.assertEqual(flow_at_test.udps.mean_clumps_size,1)
        self.assertEqual(flow_at_test.udps.stddev_clumps_size,0)
        self.assertEqual(flow_at_test.udps.skewness_clumps_size,0)
        self.assertEqual(flow_at_test.udps.variance_clumps_size,0)
        self.assertEqual(flow_at_test.udps.kurtosis_clumps_size,0)

        self.assertEqual(flow_at_test.udps.max_clumps_bytes_per_packet,229)
        self.assertEqual(flow_at_test.udps.min_clumps_bytes_per_packet,98)
        self.assertEqual(flow_at_test.udps.mean_clumps_bytes_per_packet,163.5)
        self.assertEqual(flow_at_test.udps.stddev_clumps_bytes_per_packet,70.02244538105519)
        self.assertEqual(flow_at_test.udps.skewness_clumps_bytes_per_packet,0)
        self.assertEqual(flow_at_test.udps.variance_clumps_bytes_per_packet,4903.142857142857)
        self.assertEqual(flow_at_test.udps.kurtosis_clumps_bytes_per_packet,-2.0)

        self.assertEqual(flow_at_test.udps.max_clumps_interarrival_time, 44952.0)
        self.assertEqual(flow_at_test.udps.min_clumps_interarrival_time, 0.0)
        self.assertEqual(flow_at_test.udps.mean_clumps_interarrival_time, 16886.25)
        self.assertEqual(flow_at_test.udps.stddev_clumps_interarrival_time, 23214.190461562823)
        self.assertEqual(flow_at_test.udps.skewness_clumps_interarrival_time, 0.516394118239571)
        self.assertEqual(flow_at_test.udps.variance_clumps_interarrival_time, 538898638.7857143)
        self.assertEqual(flow_at_test.udps.kurtosis_clumps_interarrival_time, -1.733322411845363)

        self.assertEqual(flow_at_test.udps.max_clump_len, [98.0,229.0,98.0,229.0,98.0,229.0,98.0,229.0])
        self.assertEqual(flow_at_test.udps.min_clump_len, [98.0,229.0,98.0,229.0,98.0,229.0,98.0,229.0])
        self.assertEqual(flow_at_test.udps.mean_clump_len, [98.0,229.0,98.0,229.0,98.0,229.0,98.0,229.0])
        self.assertEqual(flow_at_test.udps.stddev_clump_len, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.skewness_clump_len, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.variance_clump_len, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.kurtosis_clump_len, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])

        self.assertEqual(flow_at_test.udps.max_clump_interarrival_time, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.min_clump_interarrival_time, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.mean_clump_interarrival_time, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.stddev_clump_interarrival_time, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.skewness_clump_interarrival_time, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.variance_clump_interarrival_time, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.kurtosis_clump_interarrival_time, [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0])

    def test_clump_flow_2(self):
        # Given
        pcap_filepath = os.path.join(pcaps_dir, 'tls_small_pkt_payload_ratio_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=self.plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        
        #src2dst
        self.assertEqual(flow_at_test.udps.src2dst_max_clumps_len, 828.0)
        self.assertEqual(flow_at_test.udps.src2dst_min_clumps_len, 66.0)
        self.assertEqual(flow_at_test.udps.src2dst_mean_clumps_len, 503.5)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clumps_len, 322.2530061923395)
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clumps_len, -0.5495140751181137)
        self.assertEqual(flow_at_test.udps.src2dst_variance_clumps_len, 103847.0)
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clumps_len, -1.0241700487087126)

        self.assertEqual(flow_at_test.udps.src2dst_max_clumps_size, 4.0)
        self.assertEqual(flow_at_test.udps.src2dst_min_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.src2dst_mean_clumps_size, 2.25)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clumps_size, 1.2583057392117916)
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clumps_size, 0.6520236646847543)
        self.assertEqual(flow_at_test.udps.src2dst_variance_clumps_size, 1.583333333333333)
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clumps_size, -0.9030470914127426)

        self.assertEqual(flow_at_test.udps.src2dst_max_clumps_bytes_per_packet, 414.0)
        self.assertEqual(flow_at_test.udps.src2dst_min_clumps_bytes_per_packet, 66.0)
        self.assertEqual(flow_at_test.udps.src2dst_mean_clumps_bytes_per_packet, 229.0625)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clumps_bytes_per_packet, 162.1149765598478)
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clumps_bytes_per_packet, 0.12668917827875245)
        self.assertEqual(flow_at_test.udps.src2dst_variance_clumps_bytes_per_packet, 26281.265625000004)
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clumps_bytes_per_packet, -1.6820327141898344)

        self.assertEqual(flow_at_test.udps.src2dst_max_clumps_interarrival_time, 9.0)
        self.assertEqual(flow_at_test.udps.src2dst_min_clumps_interarrival_time, 0.0)
        self.assertEqual(flow_at_test.udps.src2dst_mean_clumps_interarrival_time, 2.25)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clumps_interarrival_time, 4.5)
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clumps_interarrival_time, 1.1547005383792515)
        self.assertEqual(flow_at_test.udps.src2dst_variance_clumps_interarrival_time, 20.25)
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clumps_interarrival_time, -0.6666666666666665)

        self.assertEqual(flow_at_test.udps.src2dst_max_clump_len, [66.0,571.0,723.0,173.0])
        self.assertEqual(flow_at_test.udps.src2dst_min_clump_len, [66.0,54.0,105.0,104.0])
        self.assertEqual(flow_at_test.udps.src2dst_mean_clump_len, [66.0,312.5,414.0,123.75])
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clump_len, [0.0,365.57420587344507,436.9919907732864,32.92795165205391])
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clump_len, [0.0,0.0,0.0,1.1346721570685077])
        self.assertEqual(flow_at_test.udps.src2dst_variance_clump_len, [0.0,133644.5,190962.0,1084.25])
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clump_len, [0.0,-2.0,-2.0,-0.6814517653222136])

        self.assertEqual(flow_at_test.udps.src2dst_max_clump_interarrival_time, [0.0,4.0,2.0,1.0])
        self.assertEqual(flow_at_test.udps.src2dst_min_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.src2dst_mean_clump_interarrival_time, [0.0,2.0,1.0,0.25])
        self.assertEqual(flow_at_test.udps.src2dst_stddev_clump_interarrival_time, [0.0,2.8284271247461903,1.4142135623730951,0.5])
        self.assertEqual(flow_at_test.udps.src2dst_skewness_clump_interarrival_time, [0.0,0.0,0.0,1.1547005383792515])
        self.assertEqual(flow_at_test.udps.src2dst_variance_clump_interarrival_time, [0.0,8.0,2.0,0.25])
        self.assertEqual(flow_at_test.udps.src2dst_kurtosis_clump_interarrival_time, [0.0,-2.0,-2.0,-0.6666666666666665])

        
        #dst2src
        self.assertEqual(flow_at_test.udps.dst2src_max_clumps_len, 320.0)
        self.assertEqual(flow_at_test.udps.dst2src_min_clumps_len, 66.0)
        self.assertEqual(flow_at_test.udps.dst2src_mean_clumps_len, 172.25)
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clumps_len, 112.19737073568169)
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clumps_len, 0.48303687368319326)
        self.assertEqual(flow_at_test.udps.dst2src_variance_clumps_len, 12588.25)
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clumps_len, -1.2574291459307467)

        self.assertEqual(flow_at_test.udps.dst2src_max_clumps_size, 2.0)
        self.assertEqual(flow_at_test.udps.dst2src_min_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.dst2src_mean_clumps_size, 1.5)
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clumps_size, 0.5773502691896257)
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clumps_size,  5.551115123125783e-17)
        self.assertEqual(flow_at_test.udps.dst2src_variance_clumps_size, 0.3333333333333333)
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clumps_size, -2.0)

        self.assertEqual(flow_at_test.udps.dst2src_max_clumps_bytes_per_packet, 195.0)
        self.assertEqual(flow_at_test.udps.dst2src_min_clumps_bytes_per_packet, 54.0)
        self.assertEqual(flow_at_test.udps.dst2src_mean_clumps_bytes_per_packet, 118.75)
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clumps_bytes_per_packet, 69.5)
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clumps_bytes_per_packet, 0.10922411230601688)
        self.assertEqual(flow_at_test.udps.dst2src_variance_clumps_bytes_per_packet, 4830.25)
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clumps_bytes_per_packet, -1.8185859617430495)

        self.assertEqual(flow_at_test.udps.dst2src_max_clumps_interarrival_time, 150.0)
        self.assertEqual(flow_at_test.udps.dst2src_min_clumps_interarrival_time, 145.0)
        self.assertEqual(flow_at_test.udps.dst2src_mean_clumps_interarrival_time, 147.25)
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clumps_interarrival_time, 2.217355782608343)
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clumps_interarrival_time, 0.2780305556539614)
        self.assertEqual(flow_at_test.udps.dst2src_variance_clumps_interarrival_time, 4.916666666666658)
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clumps_interarrival_time, -1.4266015512783659)

        self.assertEqual(flow_at_test.udps.dst2src_max_clump_len, [66.0,195.0,266.0,54.0])
        self.assertEqual(flow_at_test.udps.dst2src_min_clump_len, [66.0,195.0,54.0,54.0])
        self.assertEqual(flow_at_test.udps.dst2src_mean_clump_len, [66.0,195.0,160.0,54.0])
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clump_len, [0.0,0.0,149.90663761154806,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clump_len, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_variance_clump_len, [0.0,0.0,22472.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clump_len, [0.0,0.0,-2.0,0.0])

        self.assertEqual(flow_at_test.udps.dst2src_max_clump_interarrival_time, [0.0,0.0,1.0,4.0])
        self.assertEqual(flow_at_test.udps.dst2src_min_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_mean_clump_interarrival_time, [0.0,0.0,0.5,2.0])
        self.assertEqual(flow_at_test.udps.dst2src_stddev_clump_interarrival_time, [0.0,0.0,0.7071067811865476,2.8284271247461903])
        self.assertEqual(flow_at_test.udps.dst2src_skewness_clump_interarrival_time, [0.0,0.0,0.0,0.0])
        self.assertEqual(flow_at_test.udps.dst2src_variance_clump_interarrival_time, [0.0,0.0,0.5,8.0])
        self.assertEqual(flow_at_test.udps.dst2src_kurtosis_clump_interarrival_time, [0.0,0.0,-2.0,-2.0])

        #bidirectional
        self.assertEqual(flow_at_test.udps.max_clumps_len, 828.0)
        self.assertEqual(flow_at_test.udps.min_clumps_len, 66.0)
        self.assertEqual(flow_at_test.udps.mean_clumps_len, 337.875)
        self.assertEqual(flow_at_test.udps.stddev_clumps_len, 285.0460802346576)
        self.assertEqual(flow_at_test.udps.skewness_clumps_len, 0.5936481089832953)
        self.assertEqual(flow_at_test.udps.variance_clumps_len, 81251.26785714286)
        self.assertEqual(flow_at_test.udps.kurtosis_clumps_len, -1.0401259818440216)

        self.assertEqual(flow_at_test.udps.max_clumps_size, 4.0)
        self.assertEqual(flow_at_test.udps.min_clumps_size, 1)
        self.assertEqual(flow_at_test.udps.mean_clumps_size, 1.875)
        self.assertEqual(flow_at_test.udps.stddev_clumps_size, 0.991031208965115)
        self.assertEqual(flow_at_test.udps.skewness_clumps_size, 1.1914950296623115)
        self.assertEqual(flow_at_test.udps.variance_clumps_size, 0.9821428571428573)
        self.assertEqual(flow_at_test.udps.kurtosis_clumps_size, 0.7490909090909086)

        self.assertEqual(flow_at_test.udps.max_clumps_bytes_per_packet, 414.0)
        self.assertEqual(flow_at_test.udps.min_clumps_bytes_per_packet, 54.0)
        self.assertEqual(flow_at_test.udps.mean_clumps_bytes_per_packet, 173.90625)
        self.assertEqual(flow_at_test.udps.stddev_clumps_bytes_per_packet, 129.65461932573015)
        self.assertEqual(flow_at_test.udps.skewness_clumps_bytes_per_packet, 0.8510770040087812)
        self.assertEqual(flow_at_test.udps.variance_clumps_bytes_per_packet, 16810.3203125)
        self.assertEqual(flow_at_test.udps.kurtosis_clumps_bytes_per_packet, -0.5871121114366313)

        self.assertEqual(flow_at_test.udps.max_clumps_interarrival_time, 150.0)
        self.assertEqual(flow_at_test.udps.min_clumps_interarrival_time, 0.0)
        self.assertEqual(flow_at_test.udps.mean_clumps_interarrival_time, 74.75000000000001)
        self.assertEqual(flow_at_test.udps.stddev_clumps_interarrival_time, 77.57530903211776)
        self.assertEqual(flow_at_test.udps.skewness_clumps_interarrival_time, -0.003180981731713864)
        self.assertEqual(flow_at_test.udps.variance_clumps_interarrival_time, 6017.928571428572)
        self.assertEqual(flow_at_test.udps.kurtosis_clumps_interarrival_time, -1.9931840217663834)

        self.assertEqual(flow_at_test.udps.max_clump_len, [66.0, 66.0, 571.0, 195.0, 723.0, 266.0, 173.0, 54.0])
        self.assertEqual(flow_at_test.udps.min_clump_len, [66.0, 66.0, 54.0, 195.0, 105.0, 54.0, 104.0, 54.0])
        self.assertEqual(flow_at_test.udps.mean_clump_len, [66.0, 66.0, 312.5, 195.0, 414.0, 160.0, 123.75, 54.0])
        self.assertEqual(flow_at_test.udps.stddev_clump_len, [0.0, 0.0, 365.57420587344507, 0.0, 436.9919907732864, 149.90663761154806, 32.92795165205391, 0.0])
        self.assertEqual(flow_at_test.udps.skewness_clump_len, [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.1346721570685077, 0.0])
        self.assertEqual(flow_at_test.udps.variance_clump_len, [0.0, 0.0, 133644.5, 0.0, 190962.0, 22472.0, 1084.25, 0.0])
        self.assertEqual(flow_at_test.udps.kurtosis_clump_len, [0.0, 0.0, -2.0, 0.0, -2.0, -2.0, -0.6814517653222136, 0.0])

        self.assertEqual(flow_at_test.udps.max_clump_interarrival_time, [0.0, 0.0, 4.0, 0.0, 2.0, 1.0, 1.0, 4.0])
        self.assertEqual(flow_at_test.udps.min_clump_interarrival_time, [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        self.assertEqual(flow_at_test.udps.mean_clump_interarrival_time, [0.0, 0.0, 2.0, 0.0, 1.0, 0.5, 0.25, 2.0])
        self.assertEqual(flow_at_test.udps.stddev_clump_interarrival_time, [0.0, 0.0, 2.8284271247461903, 0.0, 1.4142135623730951, 0.7071067811865476, 0.5, 2.8284271247461903])
        self.assertEqual(flow_at_test.udps.skewness_clump_interarrival_time, [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.1547005383792515, 0.0])
        self.assertEqual(flow_at_test.udps.variance_clump_interarrival_time, [0.0, 0.0, 8.0, 0.0, 2.0, 0.5, 0.25, 8.0])
        self.assertEqual(flow_at_test.udps.kurtosis_clump_interarrival_time, [0.0, 0.0, -2.0, 0.0, -2.0, -2.0, -0.6666666666666665, -2.0])



class TestPacketsSizeInterarrivalTimePlugin(unittest.TestCase):

    def setUp(self):
        self.plugins = [Packets_size_and_interarrival_time()]
    
    def test_packets_size_and_interarrival_time_1(self):
        # Given
        pcap_filepath = os.path.join(pcaps_dir, 'tls_pkt_rel_time_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=self.plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')

        #Packets Size
        self.assertEqual(flow_at_test.udps.packets_size_min, 98.0)
        self.assertEqual(flow_at_test.udps.packets_size_max, 229.0)
        self.assertEqual(flow_at_test.udps.packets_size_stddev, 70.02244538105519)
        self.assertEqual(flow_at_test.udps.packets_size_first_quartile, 98.0)
        self.assertEqual(flow_at_test.udps.packets_size_second_quartile, 163.5)
        self.assertEqual(flow_at_test.udps.packets_size_third_quartile, 229.0)
        self.assertEqual(flow_at_test.udps.packets_size_mean, 163.5)
        self.assertEqual(flow_at_test.udps.packets_size_median_absoulte_deviation, 65.5)
        self.assertEqual(flow_at_test.udps.packets_size_variance, 4903.142857142857)
        self.assertEqual(flow_at_test.udps.packets_size_skewness, 0.0)
        self.assertEqual(flow_at_test.udps.packets_size_kurtosis, -2.0)
        self.assertEqual(flow_at_test.udps.packets_size_sum, 1308.0)
        
        #Packets Interarrival Time 
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_min, 0.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_max, 44952.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_stddev, 23214.190461562823)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_first_quartile, 64.5)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_second_quartile, 100.5)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_third_quartile, 44904.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_mean, 16886.25)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_median_absoulte_deviation, 69.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_variance, 538898638.7857143)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_skewness, 0.516394118239571)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_kurtosis, -1.733322411845363)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_sum, 135090.0)

    def test_packets_size_and_interarrival_time_2(self):
        # Given
        pcap_filepath = os.path.join(pcaps_dir, 'tls_small_pkt_payload_ratio_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=self.plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')

        #Packets Size
        self.assertEqual(flow_at_test.udps.packets_size_min, 54.0)
        self.assertEqual(flow_at_test.udps.packets_size_max, 723.0)
        self.assertEqual(flow_at_test.udps.packets_size_stddev, 201.13613016348617)
        self.assertEqual(flow_at_test.udps.packets_size_first_quartile, 60.0)
        self.assertEqual(flow_at_test.udps.packets_size_second_quartile, 105.0)
        self.assertEqual(flow_at_test.udps.packets_size_third_quartile, 184.0)
        self.assertEqual(flow_at_test.udps.packets_size_mean, 180.20000000000002)
        self.assertEqual(flow_at_test.udps.packets_size_median_absoulte_deviation, 51.0)
        self.assertEqual(flow_at_test.udps.packets_size_variance, 40455.74285714285)
        self.assertEqual(flow_at_test.udps.packets_size_skewness, 1.8866361368697218)
        self.assertEqual(flow_at_test.udps.packets_size_kurtosis, 2.2211745099976588)
        self.assertEqual(flow_at_test.udps.packets_size_sum, 2703)
        
        #Packets Interarrival Time 
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_min, 0.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_max, 150.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_stddev, 66.57827473483124)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_first_quartile, 0.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_second_quartile, 2.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_third_quartile, 77.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_mean, 40.666666666666664)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_median_absoulte_deviation, 2.0)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_variance, 4432.666666666667)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_skewness, 1.051871352384891)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_kurtosis, -0.8862335514542248)
        self.assertEqual(flow_at_test.udps.packets_interarrival_time_sum, 610.0)



class TestDNSCounter(unittest.TestCase):
    '''
    dns_packets   
    dns_queries   
    dns_responses 
    dns_qd_count  
    dns_an_count
    dns_ns_count
    dns_ar_count
    dns_response_digit_count
    dns_response_alpha_count
    dns_response_hypens_coun 
    dns_response_dots_count
    dns_response_ip_count
    '''
        
    def test_dns_counting_1(self):
        # TODO: read a file with a single flow, as multiprocessing by n_meters can reorder the flows
        # Given
        plugins = [DNSCounter()]
        pcap_filepath = os.path.join(pcaps_dir, 'dns_1.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins, n_meters=1)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        # bidirectional
        self.assertEqual(flow_at_test.udps.bidirectional_dns_packets,   flow_at_test.bidirectional_packets)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_queries,   1)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_responses, 1)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_qd_count,  2)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_an_count,  3)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_ns_count,  0)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_ar_count,  0)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_response_digit_count,  1)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_response_alpha_count,  93)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_response_hypens_count, 8)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_response_dots_count,   9)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_response_ip_count,     1)
        self.assertEqual(flow_at_test.udps.bidirectional_dns_response_ttls_s, [3158, 23, 3]) # in seconds
        self.assertEqual(flow_at_test.udps.bidirectional_median_dns_response_ttls_s, 23)
        self.assertEqual(flow_at_test.udps.bidirectional_mean_dns_response_digit_count, 1)
        self.assertEqual(flow_at_test.udps.bidirectional_mean_dns_response_alpha_count, 93)
        self.assertEqual(flow_at_test.udps.bidirectional_mean_dns_response_hypens_count, 8)
        self.assertEqual(flow_at_test.udps.bidirectional_mean_dns_response_dots_count, 9)
        self.assertEqual(flow_at_test.udps.bidirectional_mean_dns_response_ip_count, 1)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_mean_dns_response_ttls_s, 1061.333, 3)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_stddev_dns_response_ttls_s, 1482.5897, 3)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_variance_dns_response_ttls_s, 2198072.2222, 3)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_coeff_of_var_dns_response_ttls_s, 1.3969, 3)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_skew_from_median_dns_response_ttls_s, 2.101, 3)
        self.assertEqual(flow_at_test.udps.bidirectional_min_dns_response_ttls_s, 3)
        self.assertEqual(flow_at_test.udps.bidirectional_max_dns_response_ttls_s, 3158)
        # src -> dst
        self.assertEqual(flow_at_test.udps.src2dst_dns_queries,   1)
        self.assertEqual(flow_at_test.udps.src2dst_dns_responses, 0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_qd_count,  1)
        self.assertEqual(flow_at_test.udps.src2dst_dns_an_count,  0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_ns_count,  0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_ar_count,  0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_response_digit_count,  0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_response_alpha_count,  0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_response_hypens_count, 0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_response_dots_count,   0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_response_ip_count,     0)
        self.assertEqual(flow_at_test.udps.src2dst_dns_response_ttls_s, None) # in seconds
        self.assertEqual(flow_at_test.udps.src2dst_mean_dns_response_ttls_s, None)
        self.assertEqual(flow_at_test.udps.src2dst_stddev_dns_response_ttls_s, None)
        self.assertEqual(flow_at_test.udps.src2dst_variance_dns_response_ttls_s, None)
        self.assertEqual(flow_at_test.udps.src2dst_coeff_of_var_dns_response_ttls_s, None)
        self.assertEqual(flow_at_test.udps.src2dst_skew_from_median_dns_response_ttls_s, None)
        self.assertEqual(flow_at_test.udps.src2dst_min_dns_response_ttls_s, None)
        self.assertEqual(flow_at_test.udps.src2dst_max_dns_response_ttls_s, None)
        # dst -> src
        self.assertEqual(flow_at_test.udps.dst2src_dns_queries,   0)
        self.assertEqual(flow_at_test.udps.dst2src_dns_responses, 1)
        self.assertEqual(flow_at_test.udps.dst2src_dns_qd_count,  1)
        self.assertEqual(flow_at_test.udps.dst2src_dns_an_count,  3)
        self.assertEqual(flow_at_test.udps.dst2src_dns_ns_count,  0)
        self.assertEqual(flow_at_test.udps.dst2src_dns_ar_count,  0)
        self.assertEqual(flow_at_test.udps.dst2src_dns_response_digit_count,  1)
        self.assertEqual(flow_at_test.udps.dst2src_dns_response_alpha_count,  93)
        self.assertEqual(flow_at_test.udps.dst2src_dns_response_hypens_count, 8)
        self.assertEqual(flow_at_test.udps.dst2src_dns_response_dots_count,   9)
        self.assertEqual(flow_at_test.udps.dst2src_dns_response_ip_count,     1)
        self.assertEqual(flow_at_test.udps.dst2src_dns_response_ttls_s, [3158, 23, 3]) # in seconds
        self.assertAlmostEqual(flow_at_test.udps.dst2src_mean_dns_response_ttls_s, 1061.333, 3)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_stddev_dns_response_ttls_s, 1482.5897, 3)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_variance_dns_response_ttls_s, 2198072.2222, 3)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_coeff_of_var_dns_response_ttls_s, 1.3969, 3)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_skew_from_median_dns_response_ttls_s, 2.101, 3)
        self.assertEqual(flow_at_test.udps.dst2src_min_dns_response_ttls_s, 3)
        self.assertEqual(flow_at_test.udps.dst2src_max_dns_response_ttls_s, 3158)



class TestMostFreqPayloadRatio(unittest.TestCase):

    def test_most_freq_payload_ratio_1(self):
        # Given
        plugins = [MostFreqPayloadLenRatio()]
        pcap_filepath = os.path.join(pcaps_dir, 'http_most_freq_payload_ratio_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.src2dst_most_freq_payload_ratio, 
                        3/5)
        self.assertEqual(flow_at_test.udps.dst2src_most_freq_payload_ratio, 
                        3/4)
        self.assertEqual(flow_at_test.udps.dst2src_most_freq_payload_len, 
                        0)
        self.assertEqual(flow_at_test.udps.src2dst_most_freq_payload_len, 
                        0)



class TestSmallPacketPayloadRatio(unittest.TestCase):

    def test_small_pkt_payload_ratio_1(self):
        # Given
        plugins = [SmallPacketPayloadRatio()]
        pcap_filepath = os.path.join(pcaps_dir, 'tls_small_pkt_payload_ratio_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(flow_at_test.udps.src2dst_small_packet_payload_ratio, 
                        2/9)
        self.assertEqual(flow_at_test.udps.src2dst_small_packet_payload_packets, 
                        2)
        self.assertEqual(flow_at_test.udps.dst2src_small_packet_payload_ratio, 
                        4/6)
        self.assertEqual(flow_at_test.udps.dst2src_small_packet_payload_packets, 
                        4)



class TestPacketRelativeTime(unittest.TestCase):

    def test_pkt_rel_time_1(self):
        # Given
        plugins = [PacketRelativeTime()]
        pcap_filepath = os.path.join(pcaps_dir, 'tls_pkt_rel_time_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        # bidirectional
        self.assertEqual(flow_at_test.udps.bidirectional_packet_relative_times, 
                        [0, 132, 44999, 45065, 90006, 90075, 135027, 135090])
        self.assertEqual(flow_at_test.udps.bidirectional_min_packet_relative_times,  0)
        self.assertEqual(flow_at_test.udps.bidirectional_max_packet_relative_times, 135090)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_mean_packet_relative_times, 67549.25)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_stddev_packet_relative_times , 50309.985012296514)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_variance_packet_relative_times, 2531094591.9375)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_coeff_of_var_packet_relative_times, 0.7447896906671)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_skew_from_median_packet_relative_times, 0.0008199167)
        # src -> dst
        self.assertEqual(flow_at_test.udps.src2dst_packet_relative_times, 
                        [0, 44999, 90006, 135027])
        self.assertEqual(flow_at_test.udps.src2dst_min_packet_relative_times,  0)
        self.assertEqual(flow_at_test.udps.src2dst_max_packet_relative_times, 135027)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_mean_packet_relative_times, 67508)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_stddev_packet_relative_times , 50321.36849788)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_variance_packet_relative_times, 2532240127.5)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_coeff_of_var_packet_relative_times, 0.745413410231)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_skew_from_median_packet_relative_times, 0.0003278925135)
        # dst -> src
        self.assertEqual(flow_at_test.udps.dst2src_packet_relative_times, 
                        [132, 45065, 90075, 135090])
        self.assertEqual(flow_at_test.udps.dst2src_min_packet_relative_times,  132)
        self.assertEqual(flow_at_test.udps.dst2src_max_packet_relative_times, 135090)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_mean_packet_relative_times, 67590.5)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_stddev_packet_relative_times , 50298.565121184)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_variance_packet_relative_times, 2529945653.25)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_coeff_of_var_packet_relative_times, 0.7441661937)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_skew_from_median_packet_relative_times, 0.0012226988)



class TestByteFrequencyPlugin(unittest.TestCase):

    def test_byte_frequency_1(self):
        # Given
        plugins = [NPacketsByteFrequency(n_first_packets=1)]
        pcap_filepath = os.path.join(pcaps_dir, 'tls_pkt_rel_time_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        
        self.assertEqual(flow_at_test.udps.n_packets_byte_frequency_value, 1)
        self.assertSequenceEqual(list(flow_at_test.udps.bidirectional_n_packets_byte_frequency), 
            [
                1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1,
                1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1,
                0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0,
                0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0,
                1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0
            ]
        )
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_mean_n_packets_byte_distribution, 106.52380952) 
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_stddev_n_packets_byte_distribution, 76.95594640) 
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_median_n_packets_byte_distribution, 97)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_variance_n_packets_byte_distribution, 5922.217687074)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_coeff_of_var_n_packets_byte_distribution, 0.7224295371)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_skew_from_median_n_packets_byte_distribution, 0.3712699265)
        self.assertSequenceEqual(list(flow_at_test.udps.src2dst_n_packets_byte_frequency), 
            [
                1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1,
                1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1,
                0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0,
                0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0,
                1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0
            ]
        )
        self.assertAlmostEqual(flow_at_test.udps.src2dst_mean_n_packets_byte_distribution, 106.52380952) 
        self.assertAlmostEqual(flow_at_test.udps.src2dst_stddev_n_packets_byte_distribution, 76.95594640) 
        self.assertAlmostEqual(flow_at_test.udps.src2dst_median_n_packets_byte_distribution, 97)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_variance_n_packets_byte_distribution, 5922.217687074)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_coeff_of_var_n_packets_byte_distribution, 0.7224295371)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_skew_from_median_n_packets_byte_distribution, 0.3712699265)
        self.assertSequenceEqual(list(flow_at_test.udps.dst2src_n_packets_byte_frequency), 
            [
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
            ]
        )
        self.assertAlmostEqual(flow_at_test.udps.dst2src_mean_n_packets_byte_distribution, 0) 
        self.assertAlmostEqual(flow_at_test.udps.dst2src_stddev_n_packets_byte_distribution, 0) 
        self.assertAlmostEqual(flow_at_test.udps.dst2src_median_n_packets_byte_distribution, 0)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_variance_n_packets_byte_distribution, 0)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_coeff_of_var_n_packets_byte_distribution, 0)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_skew_from_median_n_packets_byte_distribution, 0)
        
    def test_byte_frequency_2(self):
        # Given
        plugins = [NPacketsByteFrequency(n_first_packets=2)]
        pcap_filepath = os.path.join(pcaps_dir, 'tls_pkt_rel_time_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        
        self.assertEqual(flow_at_test.udps.n_packets_byte_frequency_value, 2)
        self.assertSequenceEqual(list(flow_at_test.udps.bidirectional_n_packets_byte_frequency), 
            [
                2, 2, 2, 2, 1, 2, 2, 1, 2, 1, 2, 1, 1, 1, 2, 2, 1, 0, 1, 1, 1, 2,
                0, 2, 2, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 2, 0, 2, 1, 0, 1, 1, 1,
                1, 0, 0, 2, 2, 1, 0, 2, 0, 0, 0, 0, 1, 0, 2, 1, 0, 1, 1, 0, 1, 1,
                0, 1, 0, 2, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 2, 0, 1, 1, 2, 2, 0, 2,
                1, 1, 1, 0, 0, 0, 1, 2, 1, 2, 0, 1, 2, 1, 0, 2, 0, 2, 2, 0, 1, 0,
                1, 2, 1, 1, 0, 2, 1, 0, 0, 1, 1, 1, 1, 0, 2, 2, 1, 0, 2, 1, 1, 0,
                2, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 1, 2, 1, 2, 0, 1,
                1, 0, 1, 0, 0, 2, 0, 1, 0, 0, 1, 0, 1, 0, 2, 0, 1, 0, 1, 0, 1, 0,
                0, 1, 0, 1, 1, 2, 1, 0, 0, 0, 2, 2, 0, 2, 1, 0, 2, 1, 0, 0, 0, 0,
                1, 0, 0, 1, 1, 1, 1, 1, 2, 0, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0,
                0, 1, 1, 1, 0, 2, 0, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 2, 1, 0, 0, 0,
                1, 0, 0, 1, 1, 2, 1, 1, 2, 1, 1, 0, 0, 1
            ]
        )
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_mean_n_packets_byte_distribution, 116.31707317) 
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_stddev_n_packets_byte_distribution, 77.0165769) 
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_median_n_packets_byte_distribution, 111)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_variance_n_packets_byte_distribution, 5931.55312314)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_coeff_of_var_n_packets_byte_distribution, 0.66212615)
        self.assertAlmostEqual(flow_at_test.udps.bidirectional_skew_from_median_n_packets_byte_distribution, 0.207114106)
        self.assertSequenceEqual(list(flow_at_test.udps.src2dst_n_packets_byte_frequency), 
            [
                1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1,
                1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1,
                0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0,
                0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0,
                1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0,
                0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0
            ]
        )
        self.assertAlmostEqual(flow_at_test.udps.src2dst_mean_n_packets_byte_distribution, 106.52380952) 
        self.assertAlmostEqual(flow_at_test.udps.src2dst_stddev_n_packets_byte_distribution, 76.95594640) 
        self.assertAlmostEqual(flow_at_test.udps.src2dst_median_n_packets_byte_distribution, 97)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_variance_n_packets_byte_distribution, 5922.217687074)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_coeff_of_var_n_packets_byte_distribution, 0.7224295371)
        self.assertAlmostEqual(flow_at_test.udps.src2dst_skew_from_median_n_packets_byte_distribution, 0.3712699265)
        self.assertSequenceEqual(list(flow_at_test.udps.dst2src_n_packets_byte_frequency), 
            [
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1,
                1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1,
                1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 1, 1, 0,
                0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1,
                1, 1, 0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0, 1,
                1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0,
                1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0,
                1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0,
                1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1,
                1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0,
                0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1
            ]
        )
        self.assertAlmostEqual(flow_at_test.udps.dst2src_mean_n_packets_byte_distribution, 120.66197183) 
        self.assertAlmostEqual(flow_at_test.udps.dst2src_stddev_n_packets_byte_distribution, 76.64376046) 
        self.assertAlmostEqual(flow_at_test.udps.dst2src_median_n_packets_byte_distribution, 119)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_variance_n_packets_byte_distribution, 5874.2660186)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_coeff_of_var_n_packets_byte_distribution, 0.63519399944)
        self.assertAlmostEqual(flow_at_test.udps.dst2src_skew_from_median_n_packets_byte_distribution, 0.06505311668)
        
            

class TestReqResDiffTimePlugin(unittest.TestCase):
    def setUp(self):
        self.plugins = [ResReqDiffTime()]

    def test_stream_req_res_diff_time_1(self):
        # Given
        pcap_filepath = os.path.join(pcaps_dir, 'tls_req_res_diff_time_single.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=self.plugins)
        # When
        flows = list(streamer) # read streams/flows streamer
        # Then
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(len(flow_at_test.udps.req_res_time_diff), 43)
        self.assertSequenceEqual(flow_at_test.udps.req_res_time_diff, 
            [
                132, 44867, 66, 44941, 69, 44952, 63, 
                44949, 70, 44941, 65, 44934, 63, 44949, 
                69, 44937, 67, 44947, 68, 44924, 61, 44947, 
                63, 44949, 61, 44943, 64, 44935, 63, 44947, 
                83, 44930, 64, 44938, 65, 44938, 66, 44945, 
                66, 44940, 77, 44931, 61
            ]
        )
        avg = np.average(flow_at_test.udps.req_res_time_diff)
        self.assertAlmostEqual(avg, 21981, delta=1)
        self.assertAlmostEqual(flow_at_test.udps.min_req_res_time_diff, 61)
        self.assertAlmostEqual(flow_at_test.udps.max_req_res_time_diff, 44952)
        self.assertAlmostEqual(flow_at_test.udps.mean_req_res_time_diff, 21981.627906976744)
        self.assertAlmostEqual(flow_at_test.udps.median_req_res_time_diff, 132)
        self.assertAlmostEqual(flow_at_test.udps.stddev_req_res_time_diff, 22427.9231825)
        self.assertAlmostEqual(flow_at_test.udps.variance_req_res_time_diff, 503011738.28015137)
        self.assertAlmostEqual(flow_at_test.udps.coeff_of_var_req_res_time_diff, 1.02030310391)
        self.assertAlmostEqual(flow_at_test.udps.skew_from_median_req_res_time_diff, 2.92264616690)
            
    def test_stream_req_res_diff_time_2(self):
        # Given / Arrange
        pcap_filepath = os.path.join(pcaps_dir, 'tls_req_res_diff_time_single_2.pcap')
        streamer = NFStreamer(source=pcap_filepath, udps=self.plugins)
        # When / Act
        flows = list(streamer) # read streams/flows streamer
        # Then / Assert
        flow_at_test = flows[0]
        self.assertLessEqual(len(flows), 1, 'The PCAP test file should contain only a single stream, but contains more.')
        self.assertEqual(len(flow_at_test.udps.req_res_time_diff), 7)
        self.assertSequenceEqual(flow_at_test.udps.req_res_time_diff, [87, 5, 81, 3132, 89, 267, 94])
        avg = np.average(flow_at_test.udps.req_res_time_diff)
        self.assertAlmostEqual(avg, 536, delta=0.5)
        self.assertAlmostEqual(flow_at_test.udps.min_req_res_time_diff, 5)
        self.assertAlmostEqual(flow_at_test.udps.max_req_res_time_diff, 3132)
        self.assertAlmostEqual(flow_at_test.udps.mean_req_res_time_diff, 536.4285714)
        self.assertAlmostEqual(flow_at_test.udps.median_req_res_time_diff, 89)
        self.assertAlmostEqual(flow_at_test.udps.stddev_req_res_time_diff, 1062.163271972)
        self.assertAlmostEqual(flow_at_test.udps.variance_req_res_time_diff, 1128190.8163265)
        self.assertAlmostEqual(flow_at_test.udps.coeff_of_var_req_res_time_diff, 1.98006468809)
        self.assertAlmostEqual(flow_at_test.udps.skew_from_median_req_res_time_diff, 1.2637282324)




if __name__ == '__main__':
    unittest.main()



from EIMTC.metrics import accuracy_score


def multi_label_accuracy_score(y_true, y_pred, **kwargs):
    scores = []
    for true, pred in zip(y_true, y_pred):
        scores.append(accuracy_score(true, pred))
    
    return tuple(scores)


def multi_label_tuple_accuracy_score(y_true, y_pred):
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    score = (y_true.T == y_pred.T)[:,0]
    for i in range(1, y_true.shape[0]):
        score = score & (y_true.T == y_pred.T)[:,i]

    match_count = len(score[score == True])
    sample_count = y_true.shape[1]
    accuracy = match_count/sample_count
    
    return accuracy


from .stats import IterableStats
import pandas as pd
import numpy as np


class TLSRecordStats:
    def __init__(self, bidir_lengths):
        bidir_lengths['tls.record.length'] = bidir_lengths['tls.record.length'].astype('uint16')
        self.bidir_lengths = bidir_lengths['tls.record.length']
        self.bidir_len_stats   = IterableStats(self.bidir_lengths)
        self.src2dst_lengths = bidir_lengths[bidir_lengths['direction'] == 0]['tls.record.length']
        self.src2dst_len_stats = IterableStats(self.src2dst_lengths)
        self.dst2src_lengths = bidir_lengths[bidir_lengths['direction'] == 1]['tls.record.length']
        self.dst2src_len_stats = IterableStats(self.dst2src_lengths)
    
    def length_stats(self):
        return pd.concat([
            self.bidirectional_length_stats(),
            self.src2dst_length_stats(),
            self.dst2src_length_stats()
        ])
    
    def bidirectional_length_stats(self):
        return pd.Series({
            'bidirectional_tls_records': len(self.bidir_lengths),
            'bidirectional_tls_payload_bytes': sum(self.bidir_lengths),
            'bidirectional_tls_record_distinct_sizes': len(np.unique(self.bidir_lengths)),
            'bidirectional_mean_tls_record_size':             self.bidir_len_stats.average(),
            'bidirectional_median_tls_record_size':           self.bidir_len_stats.median(),
            'bidirectional_stddev_tls_record_size':           self.bidir_len_stats.std_deviation(),
            'bidirectional_variance_tls_record_size':         self.bidir_len_stats.variance(),
            'bidirectional_skew_from_median_tls_record_size': self.bidir_len_stats.skew_from_median(),
            'bidirectional_coeff_of_var_tls_record_size':     self.bidir_len_stats.coeff_of_variation(),
            'bidirectional_min_tls_record_size':              self.bidir_len_stats.min(),
            'bidirectional_max_tls_record_size':              self.bidir_len_stats.max(),
        })
    
    def src2dst_length_stats(self):
        return pd.Series({
            'src2dst_tls_records': len(self.src2dst_lengths),
            'src2dst_tls_payload_bytes': sum(self.src2dst_lengths),
            'src2dst_tls_record_distinct_sizes': len(np.unique(self.src2dst_lengths)),
            'src2dst_mean_tls_record_size':             self.src2dst_len_stats.average(),
            'src2dst_median_tls_record_size':           self.src2dst_len_stats.median(),
            'src2dst_stddev_tls_record_size':           self.src2dst_len_stats.std_deviation(),
            'src2dst_variance_tls_record_size':         self.src2dst_len_stats.variance(),
            'src2dst_skew_from_median_tls_record_size': self.src2dst_len_stats.skew_from_median(),
            'src2dst_coeff_of_var_tls_record_size':     self.src2dst_len_stats.coeff_of_variation(),
            'src2dst_min_tls_record_size':              self.src2dst_len_stats.min(),
            'src2dst_max_tls_record_size':              self.src2dst_len_stats.max(),
        })
        
    def dst2src_length_stats(self):
        return pd.Series({
            'dst2src_tls_records': len(self.dst2src_lengths),
            'dst2src_tls_payload_bytes': sum(self.dst2src_lengths),
            'dst2src_tls_record_distinct_sizes': len(np.unique(self.dst2src_lengths)),
            'dst2src_mean_tls_record_size':             self.dst2src_len_stats.average(),
            'dst2src_median_tls_record_size':           self.dst2src_len_stats.median(),
            'dst2src_stddev_tls_record_size':           self.dst2src_len_stats.std_deviation(),
            'dst2src_variance_tls_record_size':         self.dst2src_len_stats.variance(),
            'dst2src_skew_from_median_tls_record_size': self.dst2src_len_stats.skew_from_median(),
            'dst2src_coeff_of_var_tls_record_size':     self.dst2src_len_stats.coeff_of_variation(),
            'dst2src_min_tls_record_size':              self.dst2src_len_stats.min(),
            'dst2src_max_tls_record_size':              self.dst2src_len_stats.max(),
        })
    
    @classmethod
    def available_feature_names(cls):
        return TLSRecordStats(
            pd.DataFrame({
                'tls.record.length': [],
                'direction': []
            })
        ).length_stats().index.to_list()
        
        
        
class TLSRecordClumpStats:
    def __init__(self, bidir_lengths):
        bidir_clumps = bidir_lengths.groupby('clump_num')
        self.bidir_lengths = bidir_clumps['tls.record.length'].sum()
        self.bytes_length_stats = IterableStats(self.bidir_lengths)
        self.sizes = bidir_clumps.size()
        self.sizes_stats = IterableStats(self.sizes)
        # src2dst
        src2dst_clumps = bidir_lengths[bidir_lengths['direction'] == 0].groupby('clump_num')
        self.src2dst_lengths = src2dst_clumps['tls.record.length'].sum()
        self.src2dst_bytes_length_stats = IterableStats(self.src2dst_lengths)
        self.src2dst_sizes = src2dst_clumps.size()
        self.src2dst_sizes_stats = IterableStats(self.src2dst_sizes)
        # dst2src
        dst2src_clumps = bidir_lengths[bidir_lengths['direction'] == 1].groupby('clump_num')
        self.dst2src_lengths = dst2src_clumps['tls.record.length'].sum()
        self.dst2src_bytes_length_stats = IterableStats(self.dst2src_lengths)
        self.dst2src_sizes = dst2src_clumps.size()
        self.dst2src_sizes_stats = IterableStats(self.dst2src_sizes)
        
    def clump_stats(self):
        return pd.concat([
            self.bidirectional_stats(),
            self.src2dst_stats(),
            self.dst2src_stats(),
        ])
        
    def bidirectional_stats(self):
        return pd.Series({
            'bidirectional_tls_clumps': len(self.bidir_lengths),
            'bidirectional_mean_tls_clump_bytes':             self.bytes_length_stats.average(),
            'bidirectional_median_tls_clump_bytes':           self.bytes_length_stats.median(),
            'bidirectional_stddev_tls_clump_bytes':           self.bytes_length_stats.std_deviation(),
            'bidirectional_variance_tls_clump_bytes':         self.bytes_length_stats.variance(),
            'bidirectional_skew_from_median_tls_clump_bytes': self.bytes_length_stats.skew_from_median(),
            'bidirectional_coeff_of_var_tls_clump_bytes':     self.bytes_length_stats.coeff_of_variation(),
            'bidirectional_min_tls_clump_bytes':              self.bytes_length_stats.min(),
            'bidirectional_max_tls_clump_bytes':              self.bytes_length_stats.max(),
            'bidirectional_mean_tls_clump_sizes':             self.sizes_stats.average(),
            'bidirectional_median_tls_clump_sizes':           self.sizes_stats.median(),
            'bidirectional_stddev_tls_clump_sizes':           self.sizes_stats.std_deviation(),
            'bidirectional_variance_tls_clump_sizes':         self.sizes_stats.variance(),
            'bidirectional_skew_from_median_tls_clump_sizes': self.sizes_stats.skew_from_median(),
            'bidirectional_coeff_of_var_tls_clump_sizes':     self.sizes_stats.coeff_of_variation(),
            'bidirectional_min_tls_clump_sizes':              self.sizes_stats.min(),
            'bidirectional_max_tls_clump_sizes':              self.sizes_stats.max(),
        })
    
    def src2dst_stats(self):
        return pd.Series({
            'src2dst_tls_clumps': len(self.src2dst_lengths),
            'src2dst_mean_tls_clump_bytes':             self.src2dst_bytes_length_stats.average(),
            'src2dst_median_tls_clump_bytes':           self.src2dst_bytes_length_stats.median(),
            'src2dst_stddev_tls_clump_bytes':           self.src2dst_bytes_length_stats.std_deviation(),
            'src2dst_variance_tls_clump_bytes':         self.src2dst_bytes_length_stats.variance(),
            'src2dst_skew_from_median_tls_clump_bytes': self.src2dst_bytes_length_stats.skew_from_median(),
            'src2dst_coeff_of_var_tls_clump_bytes':     self.src2dst_bytes_length_stats.coeff_of_variation(),
            'src2dst_min_tls_clump_bytes':              self.src2dst_bytes_length_stats.min(),
            'src2dst_max_tls_clump_bytes':              self.src2dst_bytes_length_stats.max(),
            'src2dst_mean_tls_clump_sizes':             self.src2dst_sizes_stats.average(),
            'src2dst_median_tls_clump_sizes':           self.src2dst_sizes_stats.median(),
            'src2dst_stddev_tls_clump_sizes':           self.src2dst_sizes_stats.std_deviation(),
            'src2dst_variance_tls_clump_sizes':         self.src2dst_sizes_stats.variance(),
            'src2dst_skew_from_median_tls_clump_sizes': self.src2dst_sizes_stats.skew_from_median(),
            'src2dst_coeff_of_var_tls_clump_sizes':     self.src2dst_sizes_stats.coeff_of_variation(),
            'src2dst_min_tls_clump_sizes':              self.src2dst_sizes_stats.min(),
            'src2dst_max_tls_clump_sizes':              self.src2dst_sizes_stats.max(),
        })
    
    def dst2src_stats(self):
        return pd.Series({
            'dst2src_tls_clumps': len(self.dst2src_lengths),
            'dst2src_mean_tls_clump_bytes':             self.dst2src_bytes_length_stats.average(),
            'dst2src_median_tls_clump_bytes':           self.dst2src_bytes_length_stats.median(),
            'dst2src_stddev_tls_clump_bytes':           self.dst2src_bytes_length_stats.std_deviation(),
            'dst2src_variance_tls_clump_bytes':         self.dst2src_bytes_length_stats.variance(),
            'dst2src_skew_from_median_tls_clump_bytes': self.dst2src_bytes_length_stats.skew_from_median(),
            'dst2src_coeff_of_var_tls_clump_bytes':     self.dst2src_bytes_length_stats.coeff_of_variation(),
            'dst2src_min_tls_clump_bytes':              self.dst2src_bytes_length_stats.min(),
            'dst2src_max_tls_clump_bytes':              self.dst2src_bytes_length_stats.max(),
            'dst2src_mean_tls_clump_sizes':             self.dst2src_sizes_stats.average(),
            'dst2src_median_tls_clump_sizes':           self.dst2src_sizes_stats.median(),
            'dst2src_stddev_tls_clump_sizes':           self.dst2src_sizes_stats.std_deviation(),
            'dst2src_variance_tls_clump_sizes':         self.dst2src_sizes_stats.variance(),
            'dst2src_skew_from_median_tls_clump_sizes': self.dst2src_sizes_stats.skew_from_median(),
            'dst2src_coeff_of_var_tls_clump_sizes':     self.dst2src_sizes_stats.coeff_of_variation(),
            'dst2src_min_tls_clump_sizes':              self.dst2src_sizes_stats.min(),
            'dst2src_max_tls_clump_sizes':              self.dst2src_sizes_stats.max(),
        })
    
    @classmethod
    def available_feature_names(cls):
        return TLSRecordClumpStats(
            pd.DataFrame({
                'tls.record.length': [],
                'direction': [],
                'clump_num' : []
            })
        ).clump_stats().index.to_list()


# DL
from ._deepmal import DeepMALRawFlows
from ._custom_distiller import CustomDistiller
from ._m1cnn import M1CNN
from ._distiller import Distiller
from ._maldist import MalDist
from ._graphdapp import GraphDApp
from ._chainedgnn import ChainedGNN
# ML
from sklearn.svm import LinearSVC, LinearSVR, NuSVC, NuSVR, OneClassSVM, SVC, SVR
from sklearn.ensemble import (
    ExtraTreesClassifier, AdaBoostClassifier, AdaBoostRegressor, BaggingRegressor, 
    BaggingClassifier, ExtraTreesRegressor, GradientBoostingClassifier, 
    GradientBoostingRegressor, IsolationForest, RandomForestClassifier, 
    RandomForestRegressor, RandomTreesEmbedding, StackingClassifier, StackingRegressor,
    VotingClassifier, VotingRegressor
)
from sklearn.neighbors import (
    KNeighborsClassifier, KNeighborsRegressor, KNeighborsTransformer, 
    LocalOutlierFactor, NearestCentroid, NearestNeighbors, 
    NeighborhoodComponentsAnalysis, RadiusNeighborsClassifier, RadiusNeighborsRegressor, 
    RadiusNeighborsTransformer, KDTree, BallTree, DistanceMetric
)
from sklearn.naive_bayes import (
    BernoulliNB, CategoricalNB, ComplementNB, GaussianNB, MultinomialNB, 
)
from sklearn.tree import (
    DecisionTreeClassifier, DecisionTreeRegressor, ExtraTreeClassifier, ExtraTreeRegressor
)


__all__ = [ # Deep learning
    'DeepMALRawFlows',
    'CustomDistiller',
    'M1CNN',
    'Distiller',
    'MalDist',
    'GraphDApp',
    'ChainedGNN'
] + [ # SVM 
    'LinearSVC',
    'LinearSVR',
    'NuSVC', 
    'NuSVR', 
    'OneClassSVM', 
    'SVC', 
    'SVR'
] + [ # Ensemble
    'ExtraTreesClassifier',
    'AdaBoostClassifier',
    'AdaBoostRegressor,',
    'BaggingRegressor',
    'BaggingClassifier',
    'ExtraTreesRegressor',
    'GradientBoostingClassifier',
    'GradientBoostingRegressor',
    'IsolationForest',
    'RandomForestClassifier',
    'RandomForestRegressor',
    'RandomTreesEmbedding',
    'StackingClassifier',
    'StackingRegressor',
    'VotingClassifier',
    'VotingRegressor'
] + [ # Neighbors
    'KNeighborsClassifier',
    'KNeighborsRegressor',
    'KNeighborsTransformer',
    'LocalOutlierFactor',
    'NearestCentroid',
    'NearestNeighbors',
    'NeighborhoodComponentsAnalysis',
    'RadiusNeighborsClassifier',
    'RadiusNeighborsRegressor',
    'RadiusNeighborsTransformer',
    'KDTree',
    'BallTree',
    'DistanceMetric'
] + [ # Naive Nayes
    'BernoulliNB',
    'CategoricalNB',
    'ComplementNB',
    'GaussianNB',
    'MultinomialNB'
] + [ # Decision Trees
    'DecisionTreeClassifier',
    'DecisionTreeRegressor',
    'ExtraTreeClassifier',
    'ExtraTreeRegressor'
]




import subprocess
import pandas as pd
import numpy as np

def run_tshark_tls(input_filepath, output_filepath, tshark_location, error_log_filepath='./tshark_error_log.txt'):
    command =  f'{tshark_location} -r {input_filepath} -E header=y -E separator=, -E quote=d -T fields -e frame.time -e ip.src -e tcp.srcport -e ip.dst -e tcp.dstport -e ip.proto -e tls.record.version -e tls.record.length -e tls.connection_id -e tls.handshake -Y tls -2 -n'
    splitted_command = command.split()
    with open(output_filepath, "w") as outfile:
        with open(error_log_filepath, "w") as error_log_file:
            proc = subprocess.run(
                splitted_command, 
                stdout = outfile, 
                stderr = error_log_file,
                check=True
            )
    


from sklearn.metrics import classification_report, accuracy_score, recall_score, precision_score, f1_score
from ._detection import FalseAlarmRate, TruePositiveRate, DetectionRate
from ._multi_label import multi_label_accuracy_score

__all__ = [
    'FalseAlarmRate',
    'TruePositiveRate',
    'DetectionRate'
] + [ # multi label
    'multi_label_accuracy_score',
] + [ # from sklearn
    'classification_report',
    'accuracy_score',
    'recall_score',
    'precision_score',
    'f1_score'
]


from EIMTC.stats.stats import IterableStats
from nfstream import NFPlugin


class PacketRelativeTime(NFPlugin):
    ''' PacketRelativeTime |
    Relative times to the beginning of the flow. This plugin extracts statistics regarding
    relatve times of packets. 
    
    Feature Outputs:
        - packet_relative_times (list): The relative time of each packet.
        - Statistics for the above list.
        
        Name Format:
            - udps.[DIR]_[STAT]_packet_relative_times
        
        Direction (DIR) Names:
            - src2dst
            - dst2src
            - bidirectional
        
        Statistics (STAT) Names:
            - max
            - min
            - mean
            - stddev
            - variance
            - coeff_of_var
            - skew_from_median
    '''
    def __init__(self, **kwargs):
        ''' '''
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        ''' '''
        # bidirectional
        flow.udps.bidirectional_packet_relative_times = list()
        flow.udps.bidirectional_min_packet_relative_times              = 0
        flow.udps.bidirectional_max_packet_relative_times              = 0
        flow.udps.bidirectional_mean_packet_relative_times             = 0
        flow.udps.bidirectional_stddev_packet_relative_times           = None
        flow.udps.bidirectional_variance_packet_relative_times         = None
        flow.udps.bidirectional_coeff_of_var_packet_relative_times     = None
        flow.udps.bidirectional_skew_from_median_packet_relative_times = None
        # src -> dst
        flow.udps.src2dst_packet_relative_times = list()
        flow.udps.src2dst_min_packet_relative_times              = 0
        flow.udps.src2dst_max_packet_relative_times              = 0
        flow.udps.src2dst_mean_packet_relative_times             = 0
        flow.udps.src2dst_stddev_packet_relative_times           = None
        flow.udps.src2dst_variance_packet_relative_times         = None
        flow.udps.src2dst_coeff_of_var_packet_relative_times     = None
        flow.udps.src2dst_skew_from_median_packet_relative_times = None
        # dst -> src
        flow.udps.dst2src_packet_relative_times = list()
        flow.udps.dst2src_min_packet_relative_times              = 0
        flow.udps.dst2src_max_packet_relative_times              = 0
        flow.udps.dst2src_mean_packet_relative_times             = 0
        flow.udps.dst2src_stddev_packet_relative_times           = None
        flow.udps.dst2src_variance_packet_relative_times         = None
        flow.udps.dst2src_coeff_of_var_packet_relative_times     = None
        flow.udps.dst2src_skew_from_median_packet_relative_times = None

        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        ''' '''
        if packet.direction == 0: # src -> dst
            flow.udps.src2dst_packet_relative_times.append(packet.time
                                    - flow.bidirectional_first_seen_ms)
        elif packet.direction == 1:
            flow.udps.dst2src_packet_relative_times.append(packet.time
                                    - flow.bidirectional_first_seen_ms)
            
        flow.udps.bidirectional_packet_relative_times.append(packet.time
                                    - flow.bidirectional_first_seen_ms)

        
    def on_expire(self, flow):
        ''' '''
        # bidirectional
        stats = IterableStats(flow.udps.bidirectional_packet_relative_times)
        flow.udps.bidirectional_min_packet_relative_times              = stats.min()
        flow.udps.bidirectional_max_packet_relative_times              = stats.max()
        flow.udps.bidirectional_mean_packet_relative_times             = stats.average()
        flow.udps.bidirectional_stddev_packet_relative_times           = stats.std_deviation()
        flow.udps.bidirectional_variance_packet_relative_times         = stats.variance()
        flow.udps.bidirectional_coeff_of_var_packet_relative_times     = stats.coeff_of_variation()
        flow.udps.bidirectional_skew_from_median_packet_relative_times = stats.skew_from_median()
        # src -> dst
        stats = IterableStats(flow.udps.src2dst_packet_relative_times)
        flow.udps.src2dst_min_packet_relative_times              = stats.min()
        flow.udps.src2dst_max_packet_relative_times              = stats.max()
        flow.udps.src2dst_mean_packet_relative_times             = stats.average()
        flow.udps.src2dst_stddev_packet_relative_times           = stats.std_deviation()
        flow.udps.src2dst_variance_packet_relative_times         = stats.variance()
        flow.udps.src2dst_coeff_of_var_packet_relative_times     = stats.coeff_of_variation()
        flow.udps.src2dst_skew_from_median_packet_relative_times = stats.skew_from_median()
        # dst -> src
        stats = IterableStats(flow.udps.dst2src_packet_relative_times)
        flow.udps.dst2src_min_packet_relative_times              = stats.min()
        flow.udps.dst2src_max_packet_relative_times              = stats.max()
        flow.udps.dst2src_mean_packet_relative_times             = stats.average()
        flow.udps.dst2src_stddev_packet_relative_times           = stats.std_deviation()
        flow.udps.dst2src_variance_packet_relative_times         = stats.variance()
        flow.udps.dst2src_coeff_of_var_packet_relative_times     = stats.coeff_of_variation()
        flow.udps.dst2src_skew_from_median_packet_relative_times = stats.skew_from_median()



import pathlib
import os

class FilenameBasedLabelling:
    '''
    General Purpose
    '''
    def __init__(self, label_name='eimtc_default_label1'):
        self.label_name = label_name

    def __call__(self, filepath):
        label = get_filename(filepath)
        return {self.label_name: label}


class DirectoryBasedLabelling:
    '''
    General Purpose
    '''
    def __init__(self, root=None, label_names=None):
        self.root = root
        self.label_names = label_names

    def __call__(self, filepath):
        if self.label_names is None:
            label_names = iter(label_name_generator())
        else:
            label_names = iter(self.label_names)
            
        labels = []
        for p in pathlib.Path(filepath).relative_to(self.root).parents:
            label = p.name
            if label != '':
                labels.append(label)
                
        return dict(
            zip(
                label_names, 
                reversed(labels)
            )
        )


def get_filename(filepath):
    '''
    General Purpose
    '''
    return pathlib.Path(filepath).stem


def label_name_generator():
    last_num = 1
    while True:
        yield 'eimtc_default_label'+str(last_num)
        last_num += 1



from nfstream import NFPlugin


class SmallPacketPayloadRatio(NFPlugin):
    ''' SmallPacketPayloadRatio | 
    The ratio of number of small packets for direction X to the the total number of packets
    in direction X, for each direction x in {src2dst, dst2src}.
    
    Feature Outputs:
        - udps.src2dst_small_packet_payload_packets: (int) The number of small packets (upload).
        - udps.src2dst_small_packet_payload_ratio: (float) The ratio of small packets (upload) out of all upload packets.
        - udps.dst2src_small_packet_payload_packets: (int) The number of small packets (download).
        - udps.dst2src_small_packet_payload_ratio: (float) The ratio of small packets (download) out of all download packets.
    
        Name Format:
            udps.[DIR]_small_packet_payload_[FEATURE]
        
        Direction (DIR) Names:
            - src2dst
            - dst2src
            
        Feature Names:
            - packets
            - ratio
    '''
    def __init__(self, small_size=32, **kwargs):
        '''
        Args:
            `small_size` (int): the size (in bytes) of the payload that a packet is considered to be small (i.e, payload size < small_size).
        '''
        super().__init__(**kwargs)
        self.small_size = small_size

    def on_init(self, packet, flow):
        ''' '''
        flow.udps.src2dst_small_packet_payload_packets = 0
        flow.udps.src2dst_small_packet_payload_ratio   = 0
        flow.udps.dst2src_small_packet_payload_packets = 0
        flow.udps.dst2src_small_packet_payload_ratio   = 0
        

        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        '''' '''
        if packet.payload_size < self.small_size:
            if packet.direction == 0: # src2dst
                flow.udps.src2dst_small_packet_payload_packets += 1
            else:
                flow.udps.dst2src_small_packet_payload_packets += 1

    def on_expire(self, flow):
        '''' '''
        if flow.src2dst_packets != 0:
            flow.udps.src2dst_small_packet_payload_ratio = (flow.udps.src2dst_small_packet_payload_packets 
                                                            / flow.src2dst_packets)
        if flow.dst2src_packets != 0:
            flow.udps.dst2src_small_packet_payload_ratio = (flow.udps.dst2src_small_packet_payload_packets
                                                            / flow.dst2src_packets)
